"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt

import pandas as pd
import pytest
from pandas.util.testing import assert_frame_equal, assert_series_equal

from gs_quant.api.gs.data import GsDataApi
from gs_quant.context_base import ContextMeta
from gs_quant.errors import MqValueError
from gs_quant.markets import MarketDataCoordinate
from gs_quant.session import GsSession, Environment
from gs_quant.target.data import MDAPIDataQuery, MarketDataVendor, DataSetEntity

test_coordinates = (
    MarketDataCoordinate(mkt_type='Prime', mkt_quoting_style='price', mkt_asset='335320934'),
    MarketDataCoordinate(mkt_type='IR', mkt_asset='USD', mkt_class='Swap', mkt_point=('2Y',)),
)

test_str_coordinates = (
    'Prime_335320934_.price',
    'IR_USD_Swap_2Y'
)

test_defn_dict = {'id': 'EXAMPLE_FROM_SLANG',
                  'name': 'Example DataSet',
                  'description': 'This is a test.',
                  'shortDescription': '',
                  'vendor': 'Goldman Sachs',
                  'dataProduct': 'TEST',
                  'entitlements': {'query': ['internal'],
                                   'view': ['internal', 'role:DataServiceView', 'role:DataServiceAdmin'],
                                   'upload': ['internal'],
                                   'admin': ['internal', 'role:DataServiceAdmin'],
                                   'edit': ['internal', 'role:DataServiceAdmin']},
                  'parameters': {'methodology': '',
                                 'coverage': '',
                                 'notes': '',
                                 'history': '',
                                 'frequency': '',
                                 'applyMarketDataEntitlements': False,
                                 'uploadDataPolicy': 'DEFAULT_POLICY',
                                 'logicalDb': 'STUDIO_DAILY',
                                 'symbolStrategy': 'ARCTIC_LINK',
                                 'immutable': False,
                                 'includeInCatalog': False,
                                 'coverageEnabled': True},
                  'dimensions': {'timeField': 'date',
                                 'transactionTimeField': 'updateTime',
                                 'symbolDimensions': ['assetId'],
                                 'nonSymbolDimensions': [{'field': 'price', 'column': 'PRICE'}],
                                 'measures': [{'field': 'updateTime', 'column': 'UPDATE_TIME'}],
                                 'entityDimension': 'assetId'},
                  'defaults': {'startSeconds': 2592000.0},
                  'createdById': '9eb7226166a44236905cae2913cfbd3c',
                  'createdTime': '2018-07-24T00:32:25.77Z',
                  'lastUpdatedById': '4ad8ebb6480d49e6b2e9eea9210685cf',
                  'lastUpdatedTime': '2019-10-24T14:20:13.653Z'}

bond_data = [
    {
        'mktType': 'Prime',
        'mktAsset': '335320934',
        'mktQuotingStyle': 'price',
        'price': 1.0139,
        'time': pd.to_datetime('2019-01-20T01:03:00Z')
    },
    {
        'mktType': 'Prime',
        'mktAsset': '335320934',
        'mktQuotingStyle': 'price',
        'price': 1.0141,
        'time': pd.to_datetime('2019-01-20T01:08:00Z')
    }
]

swap_data = [
    {
        'mktType': 'IR',
        'mktAsset': 'USD',
        'mktClass': 'Swap',
        'mktPoint': ('2Y',),
        'mktQuotingStyle': 'ATMRate',
        'ATMRate': 0.02592,
        'time': pd.to_datetime('2019-01-20T01:09:45Z')
    }
]

bond_expected_frame = pd.DataFrame(
    data={
        'time': [pd.to_datetime('2019-01-20T01:03:00Z'), pd.to_datetime('2019-01-20T01:08:00Z')],
        'mktType': ['Prime', 'Prime'],
        'mktAsset': ['335320934', '335320934'],
        'mktQuotingStyle': ['price', 'price'],
        'value': [1.0139, 1.0141]
    },
    index=pd.DatetimeIndex(['2019-01-20T01:03:00', '2019-01-20T01:08:00']),
)

swap_expected_frame = pd.DataFrame(
    data={
        'time': [pd.to_datetime('2019-01-20T01:09:45Z')],
        'mktType': ['IR'],
        'mktAsset': ['USD'],
        'mktClass': ['Swap'],
        'mktPoint': [('2Y',)],
        'mktQuotingStyle': ['ATMRate'],
        'value': [0.02592]
    },
    index=pd.DatetimeIndex(['2019-01-20T01:09:45']),
)


def test_coordinates_data(mocker):
    start = dt.datetime(2019, 1, 2, 1, 0)
    end = dt.datetime(2019, 1, 2, 1, 10)
    # mock GsSession and data response
    mocker.patch.object(GsSession.__class__, 'default_value',
                        return_value=GsSession.get(Environment.QA, 'client_id', 'secret'))
    mocker.patch.object(GsSession.current, '_post', side_effect=[{'responses': [{'data': bond_data}]},
                                                                 {'responses': [{'data': swap_data}]},
                                                                 {'responses': [{'data': bond_data},
                                                                                {'data': swap_data}]},
                                                                 {'responses': [{'data': bond_data},
                                                                                {'data': swap_data}]}
                                                                 ])

    coord_data_result = GsDataApi.coordinates_data(coordinates=test_coordinates[0], start=start, end=end)
    assert_frame_equal(coord_data_result, bond_expected_frame)

    str_coord_data_result = GsDataApi.coordinates_data(coordinates=test_str_coordinates[1], start=start, end=end)
    assert_frame_equal(str_coord_data_result, swap_expected_frame)

    coords_data_result = GsDataApi.coordinates_data(coordinates=test_coordinates, start=start, end=end,
                                                    as_multiple_dataframes=True)
    assert len(coords_data_result) == 2
    assert_frame_equal(coords_data_result[0], bond_expected_frame)
    assert_frame_equal(coords_data_result[1], swap_expected_frame)

    GsSession.current._post.reset_mock()
    str_coords_data_result = GsDataApi.coordinates_data(coordinates=test_str_coordinates, start=start, end=end,
                                                        as_multiple_dataframes=True)
    assert len(str_coords_data_result) == 2
    assert_frame_equal(str_coords_data_result[0], bond_expected_frame)
    assert_frame_equal(str_coords_data_result[1], swap_expected_frame)
    GsSession.current._post.assert_called_once_with('/data/coordinates/query',
                                                    payload=MDAPIDataQuery(market_data_coordinates=test_coordinates,
                                                                           start_time=start,
                                                                           end_time=end,
                                                                           vendor=MarketDataVendor.Goldman_Sachs,
                                                                           format="MessagePack")
                                                    )


def test_coordinate_data_series(mocker):
    start = dt.datetime(2019, 1, 2, 1, 0)
    end = dt.datetime(2019, 1, 2, 1, 10)
    # mock GsSession and data response
    mocker.patch.object(GsSession.__class__, 'default_value',
                        return_value=GsSession.get(Environment.QA, 'client_id', 'secret'))
    mocker.patch.object(GsSession.current, '_post', side_effect=[{'responses': [{'data': bond_data}]},
                                                                 {'responses': [{'data': swap_data}]},
                                                                 {'responses': [{'data': bond_data},
                                                                                {'data': swap_data}]},
                                                                 {'responses': [{'data': bond_data},
                                                                                {'data': swap_data}]}
                                                                 ])

    bond_expected_series = pd.Series(index=bond_expected_frame.index, data=bond_expected_frame.value.values)
    swap_expected_series = pd.Series(index=swap_expected_frame.index, data=swap_expected_frame.value.values)

    coord_data_result = GsDataApi.coordinates_data_series(coordinates=test_coordinates[0], start=start, end=end)
    assert_series_equal(coord_data_result, bond_expected_series)

    str_coord_data_result = GsDataApi.coordinates_data_series(coordinates=test_str_coordinates[1], start=start, end=end)
    assert_series_equal(str_coord_data_result, swap_expected_series)

    coords_data_result = GsDataApi.coordinates_data_series(coordinates=test_coordinates, start=start, end=end)
    assert len(coords_data_result) == 2
    assert_series_equal(coords_data_result[0], bond_expected_series)
    assert_series_equal(coords_data_result[1], swap_expected_series)

    GsSession.current._post.reset_mock()
    str_coords_data_result = GsDataApi.coordinates_data_series(coordinates=test_str_coordinates, start=start, end=end)
    assert len(str_coords_data_result) == 2
    assert_series_equal(str_coords_data_result[0], bond_expected_series)
    assert_series_equal(str_coords_data_result[1], swap_expected_series)
    GsSession.current._post.assert_called_with('/data/coordinates/query',
                                               payload=MDAPIDataQuery(market_data_coordinates=test_coordinates,
                                                                      start_time=start,
                                                                      end_time=end,
                                                                      vendor=MarketDataVendor.Goldman_Sachs,
                                                                      format="MessagePack")
                                               )


def test_coordinate_last(mocker):
    as_of = dt.datetime(2019, 1, 2, 1, 10)
    data = {'responses': [
        {'data': [
            {
                'mktType': 'Prime',
                'mktAsset': '335320934',
                'mktQuotingStyle': 'price',
                'price': 1.0141,
                'time': '2019-01-20T01:08:00Z'
            }
        ]},
        {'data': [
            {
                'mktType': 'IR',
                'mktAsset': 'USD',
                'mktClass': 'Swap',
                'mktPoint': ('2Y',),
                'mktQuotingStyle': 'ATMRate',
                'ATMRate': 0.02592,
                'time': '2019-01-20T01:09:45Z'
            }
        ]}
    ]}

    expected_result = pd.DataFrame(
        data={
            'mktType': ['Prime', 'IR'],
            'mktAsset': ['335320934', 'USD'],
            'mktClass': [None, 'Swap'],
            'mktPoint': [None, ('2Y',)],
            'mktQuotingStyle': ['price', None],
            'value': [1.0141, 0.02592]
        }
    )

    # mock GsSession and data response
    mocker.patch.object(GsSession.__class__, 'default_value',
                        return_value=GsSession.get(Environment.QA, 'client_id', 'secret'))
    GsSession.current._post = mocker.Mock(return_value=data)

    result = GsDataApi.coordinates_last(coordinates=test_coordinates, as_of=as_of, as_dataframe=True)
    assert result.equals(expected_result)

    GsSession.current._post.reset_mock()
    result_from_str = GsDataApi.coordinates_last(coordinates=test_str_coordinates, as_of=as_of, as_dataframe=True)
    assert result_from_str.equals(expected_result)
    GsSession.current._post.assert_called_once_with('/data/coordinates/query/last',
                                                    payload=MDAPIDataQuery(market_data_coordinates=test_coordinates,
                                                                           end_time=as_of,
                                                                           vendor=MarketDataVendor.Goldman_Sachs,
                                                                           format="MessagePack")
                                                    )


def test_get_coverage_api(mocker):
    test_coverage_data = {'results': [{'gsid': 'gsid1'}]}

    mocker.patch.object(ContextMeta, 'current', return_value=GsSession(Environment.QA))
    mocker.patch.object(ContextMeta.current, '_get', return_value=test_coverage_data)
    data = GsDataApi.get_coverage('MA_RANK')

    assert [{'gsid': 'gsid1'}] == data


def test_get_many_defns_api(mocker):
    test_defn = DataSetEntity.from_dict(test_defn_dict)
    mock_response = {'results': (test_defn,), 'totalResults': 1}

    expected_response = (test_defn,)

    # mock GsSession
    mocker.patch.object(GsSession.__class__, 'default_value',
                        return_value=GsSession.get(Environment.QA, 'client_id', 'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsDataApi.get_many_definitions()
    GsSession.current._get.assert_called_with('/data/datasets?limit=100', cls=DataSetEntity)
    assert response == expected_response


def test_coordinates_converter():
    coord = GsDataApi._coordinate_from_str("A_B_C_D")
    assert str(coord) == 'A|B|C|D|'

    coord = GsDataApi._coordinate_from_str("A_B_C.E")
    assert str(coord) == 'A|B|C||E'

    coord = GsDataApi._coordinate_from_str("A_B_.E")
    assert str(coord) == 'A|B|||E'

    coord = GsDataApi._coordinate_from_str("A_B_C_D_E.F")
    assert str(coord) == 'A|B|C|D_E|F'

    with pytest.raises(MqValueError, match='invalid coordinate A'):
        GsDataApi._coordinate_from_str("A")


def test_get_many_coordinates(mocker):
    coordinates = [
        {
            'id': 'MC123',
            'name': 'A_B_C_D_E.F1'
        },
        {
            'id': 'MC123',
            'name': 'A_B_C_D_E.F2'
        }
    ]
    mocker.patch.object(GsSession.__class__, 'default_value',
                        return_value=GsSession.get(Environment.QA, 'client_id', 'secret'))
    GsSession.current._post = mocker.Mock(return_value={'results': coordinates})
    response = GsDataApi.get_many_coordinates(mkt_type='A', mkt_asset='B')
    assert response == ('A_B_C_D_E.F1', 'A_B_C_D_E.F2')


if __name__ == "__main__":
    pytest.main(args=["test_data.py"])



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
from typing import Tuple, Union

import numpy as np

from gs_quant.data import Dataset


class GsCalendar:

    __CALENDAR_CACHE = {}

    DATE_LOW_LIMIT = dt.date(1952, 1, 1)
    DATE_HIGH_LIMIT = dt.date(2052, 12, 31)
    DEFAULT_WEEK_MASK = '1111100'  # Default to Sat, Sun weekend days

    def __init__(self, calendars: Union[str, Tuple[str, ...]] = ()):
        if isinstance(calendars, str):
            calendars = (calendars,)

        self.__calendars = calendars
        self.__holidays = set()
        self.__business_day_calendars = {}

    @staticmethod
    def get(calendars: Union[str, Tuple]):
        if isinstance(calendars, str):
            calendars = (calendars,)

        return GsCalendar.__CALENDAR_CACHE.setdefault(calendars, GsCalendar(calendars))

    @staticmethod
    def reset():
        GsCalendar.__CALENDAR_CACHE = set()

    def calendars(self) -> Tuple:
        return self.__calendars

    @property
    def holidays(self) -> set:
        if self.__calendars and not self.__holidays:
            dataset = Dataset(Dataset.GS.HOLIDAY)
            for holiday_id in self.__calendars:
                data = dataset.get_data(exchange=holiday_id, start=self.DATE_LOW_LIMIT, end=self.DATE_HIGH_LIMIT)
                if not data.empty:
                    self.__holidays.update(data.index.values.astype('datetime64[D]'))

        return self.__holidays

    def business_day_calendar(self, week_mask: str = None) -> np.busdaycalendar:
        return self.__business_day_calendars.setdefault(week_mask, np.busdaycalendar(
            weekmask=week_mask or self.DEFAULT_WEEK_MASK, holidays=tuple(self.holidays)))



# Copyright 2018 Goldman Sachs.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
#
# Chart Service will attempt to make public functions (not prefixed with _) from this module available. Such functions
# should be fully documented: docstrings should describe parameters and the return value, and provide a 1-line
# description. Type annotations should be provided for parameters.
from .statistics import *
from ..errors import *
import itertools
from dataclasses import dataclass
from typing import Union, Tuple
import numpy as np
import pandas as pd
from statsmodels.tsa.arima_model import ARIMA
from statsmodels.tools.eval_measures import mse
from gs_quant.api.gs.data import GsDataApi
from gs_quant.data import DataContext
from gs_quant.datetime.date import DayCountConvention
from gs_quant.markets.securities import Asset
from gs_quant.target.common import Currency
from gs_quant.timeseries.datetime import align

"""
Econometrics timeseries library is for standard economic and time series analytics operations, including returns,
diffs, lags, volatilities and other numerical operations which are generally finance-oriented
"""


class AnnualizationFactor(IntEnum):
    DAILY = 252
    WEEKLY = 52
    SEMI_MONTHLY = 26
    MONTHLY = 12
    QUARTERLY = 4
    ANNUALLY = 1


class SharpeAssets(Enum):
    USD = 'MAP35DA6K5B1YXGX'
    AUD = 'MAFRZWJ790MQY0EW'
    CHF = 'MAS0NN4ZX7NYXB36'
    EUR = 'MA95W0N1214395N8'
    GBP = 'MA41ZEFTWR8Q7HBM'
    JPY = 'MA8GXV3SJ0TXH1JV'
    SEK = 'MAGNZZY0GJ4TATNG'


def excess_returns(price_series: pd.Series, benchmark_or_rate: Union[Asset, Currency, float], *,
                   day_count_convention=DayCountConvention.ACTUAL_360) -> pd.Series:
    if isinstance(benchmark_or_rate, float):
        er = [price_series.iloc[0]]
        for j in range(1, len(price_series)):
            fraction = day_count_fraction(price_series.index[j - 1], price_series.index[j], day_count_convention)
            er.append(er[-1] + price_series.iloc[j] - price_series.iloc[j - 1] * (1 + benchmark_or_rate * fraction))
        return pd.Series(er, index=price_series.index)

    if isinstance(benchmark_or_rate, Currency):
        try:
            marquee_id = SharpeAssets[benchmark_or_rate.value].value
        except KeyError:
            raise MqValueError(f"unsupported currency {benchmark_or_rate}")
    else:
        marquee_id = benchmark_or_rate.get_marquee_id()

    with DataContext(price_series.index[0], price_series.index[-1]):
        q = GsDataApi.build_market_data_query([marquee_id], QueryType.SPOT)
        df = GsDataApi.get_market_data(q)
    curve, bench_curve = align(price_series, df['spot'], Interpolate.INTERSECT)

    e_returns = [curve.iloc[0]]
    for i in range(1, len(curve)):
        multiplier = 1 + curve.iloc[i] / curve.iloc[i - 1] - bench_curve.iloc[i] / bench_curve.iloc[i - 1]
        e_returns.append(e_returns[-1] * multiplier)
    return pd.Series(e_returns, index=curve.index)


def _annualized_return(levels: pd.Series, rolling: int) -> pd.Series:
    if rolling == 0:
        rolling = len(levels)

    starting = [0] * rolling
    starting.extend([a for a in range(1, len(levels) - rolling + 1)])
    points = list(
        map(lambda d, v, i: pow(v / levels[i], 365.25 / (d - levels.index[i]).days) - 1, levels.index[1:],
            levels.values[1:], starting[1:]))
    points.insert(0, 0)
    return pd.Series(points, index=levels.index)


def _get_ratio(input_series: pd.Series, benchmark_or_rate: Union[Asset, float, str], rolling: int, *,
               day_count_convention: DayCountConvention, curve_type: CurveType = CurveType.PRICES) -> pd.Series:
    if curve_type == CurveType.PRICES:
        er = excess_returns(input_series, benchmark_or_rate, day_count_convention=day_count_convention)
    else:
        assert curve_type == CurveType.EXCESS_RETURNS
        er = input_series

    ann_return = _annualized_return(er, rolling)
    ann_vol = volatility(er, rolling).iloc[1:] if rolling > 0 else volatility(er)
    return ann_return / ann_vol * 100


class RiskFreeRateCurrency(Enum):
    AUD = "AUD"
    CHF = "CFH"
    EUR = "EUR"
    GBP = "GBP"
    JPY = "JPY"
    SEK = "SEK"
    USD = "USD"


@plot_session_function
def sharpe_ratio(series: pd.Series, currency: RiskFreeRateCurrency, rolling: int = 0,
                 curve_type: CurveType = CurveType.PRICES) -> pd.Series:
    """
    Calculate Sharpe ratio

    :param series: series of prices or excess returns for an asset
    :param currency: currency for risk-free rate
    :param curve_type: whether input series is of prices or excess returns
    :param rolling: rolling window to use
    :return: Sharpe ratio

    **Usage**

    Given a price series and risk-free rate (a number, currency, or cash asset), returns the rolling Sharpe ratio.

    For a fixed rate R, excess returns E are calculated as:

    :math:`E_t = E_{t-1} + P_t - P_{t-1} * (1 + R * DCF_{t-1,t})`

    Subscripts refers to dates in the price series.

    P is a point in the price series.

    DCF is the day count fraction using the Act/360 convention.
    """
    return _get_ratio(series, Currency(currency.value), rolling, day_count_convention=DayCountConvention.ACTUAL_360,
                      curve_type=curve_type)


@plot_function
def returns(series: pd.Series, obs: int = 1, type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Calculate returns from price series

    :param series: time series of prices
    :param obs: number of observations
    :param type: returns type: simple, logarithmic or absolute
    :return: date-based time series of return

    **Usage**

    Compute returns series from price levels, based on the value of *type*:

    ===========   =============================
    Type          Description
    ===========   =============================
    simple        Simple arithmetic returns
    logarithmic   Logarithmic returns
    absolute      Absolute returns
    ===========   =============================

    *Simple*

    Simple geometric change in asset prices, which can be aggregated across assets

    :math:`Y_t = \\frac{X_t}{X_{t-obs}} - 1`

    where :math:`X_t` is the asset price at time :math:`t`

    *Logarithmic*

    Natural logarithm of asset price changes, which can be aggregated through time

    :math:`Y_t = log(X_t) - log(X_{t-obs})`

    where :math:`X_t` is the asset price at time :math:`t`

    *Absolute*

    Absolute change in asset prices

    :math:`Y_t = X_t - X_{t-obs}`

    where :math:`X_t` is the asset price at time :math:`t`

    **Examples**

    Generate price series and take compute returns

    >>> prices = generate_series(100)
    >>> returns = returns(prices)

    **See also**

    :func:`prices`
    """

    if series.size < 1:
        return series

    if type == Returns.SIMPLE:
        ret_series = series / series.shift(obs) - 1
    elif type == Returns.LOGARITHMIC:
        log_s = series.apply(math.log)
        ret_series = log_s - log_s.shift(obs)
    elif type == Returns.ABSOLUTE:
        ret_series = series - series.shift(obs)
    else:
        raise MqValueError('Unknown returns type (use simple / logarithmic / absolute)')

    return ret_series


@plot_function
def prices(series: pd.Series, initial: int = 1, type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Calculate price levels from returns series

    :param series: time series of returns
    :param initial: initial price level
    :param type: returns type: simple, logarithmic or absolute
    :return: date-based time series of return

    **Usage**

    Compute price levels from returns series, based on the value of *type*:

    ===========   =============================
    Type          Description
    ===========   =============================
    simple        Simple arithmetic returns
    logarithmic   Logarithmic returns
    absolute      Absolute returns
    ===========   =============================

    *Simple*

    Compute asset price series from simple returns:

    :math:`Y_t = (1 + X_{t-1}) Y_{t-1}`

    where :math:`X_t` is the asset price at time :math:`t` and :math:`Y_0 = initial`

    *Logarithmic*

    Compute asset price series from logarithmic returns:

    :math:`Y_t = e^{X_{t-1}} Y_{t-1}`

    where :math:`X_t` is the asset price at time :math:`t` and :math:`Y_0 = initial`

    *Absolute*

    Compute asset price series from absolute returns:

    :math:`Y_t = X_{t-1} + Y_{t-1}`

    where :math:`X_t` is the asset price at time :math:`t` and :math:`Y_0 = initial`

    **Examples**

    Generate price series and take compute returns

    >>> series = generate_series(100)
    >>> returns = prices(returns(series))

    **See also**

    :func:`returns` :func:`product` :func:`exp`
    """

    if series.size < 1:
        return series

    if type == Returns.SIMPLE:
        return product(1 + series) * initial
    elif type == Returns.LOGARITHMIC:
        return product(series.apply(math.exp)) * initial
    elif type == Returns.ABSOLUTE:
        return sum_(series) + initial
    else:
        raise MqValueError('Unknown returns type (use simple / Logarithmic / absolute)')


@plot_function
def index(x: pd.Series, initial: int = 1) -> pd.Series:
    """
    Geometric series normalization

    :param x: time series
    :param initial: initial value
    :return: normalized time series

    **Usage**

    Divides every value in x by the initial value of x:

    :math:`Y_t = initial * X_t / X_0`

    where :math:`X_0` is the first value in the series

    **Examples**

    Normalize series to 1:

    >>> series = generate_series(100)
    >>> returns = index(series)

    **See also**

    :func:`returns`

    """
    i = x.first_valid_index()
    return pd.Series() if i is None else initial * x / x[i]


@plot_function
def change(x: pd.Series) -> pd.Series:
    """
    Arithmetic series normalization

    :param x: time series
    :return: normalized time series

    **Usage**

    Compute difference of every value from the initial value of x:

    :math:`Y_t = X_t - X_0`

    where :math:`X_0` is the first value in the series

    **Examples**

    Change in level from initial value:

    >>> series = generate_series(100)
    >>> returns = change(series)

    **See also**

    :func:`index`

    """
    return x - x[0]


def _get_annualization_factor(x):
    prev_idx = x.index[0]
    distances = []

    for idx, value in x.iloc[1:].iteritems():
        d = (idx - prev_idx).days
        if d == 0:
            raise MqValueError('multiple data points on same date')
        distances.append(d)
        prev_idx = idx

    average_distance = numpy.average(distances)
    if average_distance < 2.1:
        factor = AnnualizationFactor.DAILY
    elif 6 <= average_distance < 8:
        factor = AnnualizationFactor.WEEKLY
    elif 14 <= average_distance < 17:
        factor = AnnualizationFactor.SEMI_MONTHLY
    elif 25 <= average_distance < 35:
        factor = AnnualizationFactor.MONTHLY
    elif 85 <= average_distance < 97:
        factor = AnnualizationFactor.QUARTERLY
    elif 360 <= average_distance < 386:
        factor = AnnualizationFactor.ANNUALLY
    else:
        raise MqValueError('Cannot infer annualization factor, average distance: ' + str(average_distance))
    return factor


@plot_function
def annualize(x: pd.Series) -> pd.Series:
    """
    Annualize series based on sample observation frequency

    :param x: time series of prices
    :return: date-based time series of annualized values

    **Usage**

    Based on number of days between observations, will determine an annualization factor and then adjust values
    accordingly. Useful for annualizing daily or monthly returns

    :math:`Y_t = X_t * \sqrt{F}`

    Annualization factors as follows, based on period implied by observations:

    =========   =============================
    Period      Annualization Factor (F)
    =========   =============================
    Daily       :math:`252`
    Weekly      :math:`52`
    Bi-Weekly   :math:`26`
    Monthly     :math:`12`
    Quarterly   :math:`4`
    Annually    :math:`1`
    =========   =============================

    **Examples**

    Annualize daily returns series:

    >>> prices = generate_series(100)
    >>> ann = annualize(returns(prices))

    **See also**

    :func:`returns`
    """

    factor: int = _get_annualization_factor(x)
    return x * math.sqrt(factor)


@plot_function
def volatility(x: pd.Series, w: Union[Window, int] = Window(None, 0),
               returns_type: Returns = Returns.SIMPLE) -> pd.Series:
    """
    Realized volatility of price series

    :param x: time series of prices
    :param w: Window or int: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value. Window size defaults to length of series.
    :param returns_type: returns type: simple, logarithmic or absolute
    :return: date-based time series of return

    **Usage**

    Calculate rolling annualized realized volatility of a price series over a given window. Annual volatility of 20% is
    returned as 20.0:

    :math:`Y_t = \sqrt{\\frac{1}{N-1} \sum_{i=t-w+1}^t (R_t - \overline{R_t})^2} * \sqrt{252} * 100`

    where N is the number of observations in each rolling window :math:`w`, :math:`R_t` is the return on time
    :math:`t` based on *returns_type*

    ===========   =======================================================
    Type          Description
    ===========   =======================================================
    simple        Simple geometric change in asset prices:
                  :math:`R_t = \\frac{X_t}{X_{t-1}} - 1`
                  where :math:`X_t` is the asset price at time :math:`t`
    logarithmic   Natural logarithm of asset price changes:
                  :math:`R_t = log(X_t) - log(X_{t-1})`
                  where :math:`X_t` is the asset price at time :math:`t`
    absolute      Absolute change in asset prices:
                  :math:`Y_t = X_t - X_{t-obs}`
                  where :math:`X_t` is the asset price at time :math:`t`
    ===========   =======================================================

    and :math:`\overline{R_t}` is the mean value over the same window:

    :math:`\overline{R_t} = \\frac{\sum_{i=t-w+1}^{t} R_t}{N}`

    If window is not provided, computes realized volatility over the full series

    **Examples**

    Compute rolling :math:`1` month (:math:`22` business day) annualized volatility of price series

    >>> series = generate_series(100)
    >>> vol_series = volatility(series, 22)
    >>> vol_series = volatility(series, Window(22, 30))

    **See also**

    :func:`std` :func:`annualize` :func:`returns`

    """
    w = normalize_window(x, w)

    if x.size < 1:
        return x

    return apply_ramp(annualize(std(returns(x, type=returns_type), Window(w.w, 0))).mul(100), w)


@plot_function
def correlation(x: pd.Series, y: pd.Series,
                w: Union[Window, int] = Window(None, 0), type_: SeriesType = SeriesType.PRICES) -> pd.Series:
    """
    Rolling correlation of two price series

    :param x: price series
    :param y: price series
    :param w: Window or int: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value. Window size defaults to length of series.
    :param type_: type of both input series: prices or returns
    :return: date-based time series of correlation

    **Usage**

    Calculate rolling `realized correlation <https://en.wikipedia.org/wiki/Correlation_and_dependence>`_,
    :math:`\\rho_t` of two price series over a given window:

    :math:`\\rho_t = \\frac{\sum_{i=t-w+1}^t (R_t - \overline{R_t})(Y_t - \overline{S_t})}{(N-1)\sigma R_t\sigma S_t}`

    where N is the number of observations in each rolling window, :math:`w`, and :math:`R_t` and :math:`S_t` are the
    simple returns for each series on time :math:`t`

    If prices are provided:

    :math:`R_t = \\frac{X_t}{X_{t-1}} - 1` and :math:`S_t = \\frac{Y_t}{Y_{t-1}} - 1`

    If returns are provided:

    :math:`R_t = X_t` and :math:`S_t = Y_t`

    :math:`\overline{R_t}`, :math:`\overline{S_t}` are the mean values, and :math:`\sigma R_{t}` and
    :math:`\sigma S_{t}` are the sample standard deviations, of  series
    :math:`R_t` and :math:`S_t` over the same window

    If window is not provided, computes realized correlation over the full series

    **Examples**

    Compute rolling :math:`1` month (:math:`22` business day) correlation of price series

    >>> series1 = generate_series(100)
    >>> series2 = generate_series(100)
    >>> corr = correlation(series1, series2, 22)

    **See also**

    :func:`std` :func:`returns`

    """
    w = normalize_window(x, w)

    if x.size < 1:
        return x

    given_prices = type_ == SeriesType.PRICES
    ret_1 = returns(x) if given_prices else x
    ret_2 = returns(y) if given_prices else y

    clean_ret1 = ret_1.dropna()
    clean_ret2 = ret_2.dropna()

    if isinstance(w.w, pd.DateOffset):
        values = [clean_ret1.loc[(clean_ret1.index > idx - w.w) & (clean_ret1.index <= idx)].corr(clean_ret2)
                  for idx in clean_ret1.index]
        corr = pd.Series(values, index=clean_ret1.index)
    else:
        corr = clean_ret1.rolling(w.w, 0).corr(clean_ret2)

    return apply_ramp(interpolate(corr, x, Interpolate.NAN), w)


@plot_function
def beta(x: pd.Series, b: pd.Series, w: Union[Window, int] = Window(None, 0), prices: bool = True) -> pd.Series:
    """
    Rolling beta of price series and benchmark

    :param x: time series of prices
    :param b: time series of benchmark prices
    :param w: Window or int: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value. Window size defaults to length of series.
    :param prices: True if input series are prices, False if they are returns
    :return: date-based time series of beta

    **Usage**

    Calculate rolling `beta <https://en.wikipedia.org/wiki/Beta_(finance)>`_,
    :math:`\\beta_t` of a series to a benchmark over a given window:

    :math:`R_t = \\alpha_t + \\beta S_t + \epsilon_t`

    Calculated as:

    :math:`\\beta_t = \\frac{\sum_{i=t-w+1}^t Cov(R_t, S_t)}{Var(S_t)}`

    where N is the number of observations in each rolling window, :math:`w`, and :math:`R_t` and :math:`S_t` are the
    simple returns for each series on time :math:`t`:

    :math:`R_t = \\frac{X_t}{X_{t-1}} - 1` and :math:`S_t = \\frac{b_t}{b_{t-1}} - 1`

    If prices = False, assumes returns are provided:

    :math:`R_t = X_t` and :math:`S_t = b_t`

    :math:`Cov(R_t, S_t)` and :math:`Var(S_t)` are the mean and variance of  series
    :math:`R_t` and :math:`S_t` over the same window

    If window is not provided, computes beta over the full series

    **Examples**

    Compute rolling :math:`1` month (:math:`22` business day) beta of two price series

    >>> series = generate_series(100)
    >>> benchmark = generate_series(100)
    >>> b = beta(series, benchmark, 22)

    **See also**

    :func:`var` :func:`cov` :func:`correlation` :func:`returns`
    """
    w = normalize_window(x, w)

    ret_series = returns(x) if prices else x
    ret_benchmark = returns(b) if prices else b

    if isinstance(w.w, pd.DateOffset):
        result = pd.Series([ret_series.loc[(ret_series.index > idx - w.w) & (ret_series.index <= idx)].cov(
            ret_benchmark.loc[(ret_benchmark.index > idx - w.w) & (ret_benchmark.index <= idx)]
        ) / ret_benchmark.loc[(ret_benchmark.index > idx - w.w) & (ret_benchmark.index <= idx)].var()
            for idx in ret_series.index], index=ret_series.index)
    else:
        cov = ret_series.rolling(w.w, 0).cov(ret_benchmark.rolling(w.w, 0))
        result = cov / ret_benchmark.rolling(w.w, 0).var()

    # do not compute initial values as they may be extreme when sample size is small

    result[0:3] = np.nan

    return apply_ramp(interpolate(result, x, Interpolate.NAN), w)


@plot_function
def max_drawdown(x: pd.Series, w: Union[Window, int] = Window(None, 0)) -> pd.Series:
    """
    Compute the maximum peak to trough drawdown over a rolling window.

    :param x: time series
    :param w: Window or int: size of window and ramp up to use. e.g. Window(22, 10) where 22 is the window size
              and 10 the ramp up value. Window size defaults to length of series.
    :return: time series of rolling maximum drawdown

    **Examples**

    Compute the maximum peak to trough `drawdown <https://en.wikipedia.org/wiki/Drawdown_(economics)>`_

    >>> series = generate_series(100)
    >>> max_drawdown(series)

    **See also**

    :func:`returns`

    """
    w = normalize_window(x, w)
    if isinstance(w.w, pd.DateOffset):
        scores = pd.Series([x[idx] / x.loc[(x.index > idx - w.w) & (x.index <= idx)].max() - 1 for idx in x.index],
                           index=x.index)
        result = pd.Series([scores.loc[(scores.index > idx - w.w) & (scores.index <= idx)].min()
                            for idx in scores.index], index=scores.index)
    else:
        rolling_max = x.rolling(w.w, 0).max()
        result = (x / rolling_max - 1).rolling(w.w, 0).min()
    return apply_ramp(result, w)


@dataclass
class ARIMABestParams:
    freq: str = ''
    p: int = None
    d: int = None
    q: int = None
    const: float = None
    ar_coef: List[float] = None
    ma_coef: List[float] = None
    resid: List[float] = None
    series: pd.Series = None


class Arima:
    """
    ARIMA is the Autoregressive Integrated Moving Average Model and is used
    to normalize and forecast time series data. ARIMA has 3 parameters:
    (p, d, q) where:
        :p is the number of autoregressive terms
        :d is the number of nonseasonal differences
        :q is the number of lagged forecast errors in the prediction equation
    An ARIMA model is selected from the Cartesian product of sets p, q, and d.
    The time series is split into train and test sets and an ARIMA model is fit
    for every combination on the training set. The model with the lowest
    mean-squared error (MSE) on the test set is selected as the best model. The
    original times series can then be transformed by the best model.
    **Examples**
    >>> series = generate_series(100)
    >>> ar = econometrics.Arima()
    >>> ar.fit(series, train_size=0.8)
    >>> transformed_time_series = ar.transform(series)
    """

    def __init__(self):
        self.best_params = {}

    @staticmethod
    def _evaluate_arima_model(X: Union[pd.Series, pd.DataFrame], arima_order: Tuple[int, int, int],
                              train_size: Union[float, int, None], freq: str) -> Tuple[float, dict]:
        train_size = int(len(X) * 0.75) if train_size is None else int(len(X) * train_size) \
            if isinstance(train_size, float) else train_size
        train, test = X[:train_size].astype(float), X[train_size:].astype(float)

        model = ARIMA(train, order=arima_order, freq=freq)
        model_fit = model.fit(disp=False, method='css', trend='nc')

        # calculate test error
        yhat = model_fit.forecast(len(test))[0]
        error = mse(test, yhat)

        return error, model_fit

    def fit(self, X: Union[pd.Series, pd.DataFrame], train_size: Union[float, int, None] = None,
            p_vals: list = (0, 1, 2), d_vals: list = (0, 1, 2), q_vals: list = (0, 1, 2), freq: str = None) -> 'arima':
        """
        Train a combination of ARIMA models. If pandas DataFrame, finds the
        best arima model parameters for each column. If pandas Series, finds
        the best arima model parameters for the series.
        :param X: time series to be operated on; required parameter
        :param train_size: if float, should be between 0.0 and 1.0 and
        represent the proportion of the dataset to include in the train split.
        If int, represents the absolute number of train samples. If None,
        the value is automatically set 0.75
        :p_vals: number of autoregressive terms to search; default is [0,1,2]
        :d_vals: number of differences to search; default is [0,1,2]
        :q_vals: number of lagged forecast to search; always [0,1,2]
        :freq: frequency of time series, default is None
        :return: self
        """
        if isinstance(X, pd.Series):
            X = X.to_frame()

        for series_id in X.columns:
            series = X[series_id]
            best_score = float('inf')
            best_order = None
            best_const = None
            best_ar_coef = None
            best_ma_coef = None
            best_resid = None
            for order in list(itertools.product(*[p_vals, d_vals, q_vals])):
                try:
                    error, model_fit = self._evaluate_arima_model(series, order, train_size, freq)
                    if error < best_score:
                        best_score = error
                        best_order = order
                        best_const = model_fit.params.to_dict().get('const', 0)
                        best_ar_coef = model_fit.arparams
                        best_ma_coef = model_fit.maparams
                        best_resid = model_fit.resid
                except Exception as e:
                    print('   {}'.format(e))
                    continue

            p, d, q = best_order
            self.best_params[series_id] = ARIMABestParams(freq, p, d, q, best_const, best_ar_coef, best_ma_coef,
                                                          best_resid, series)
        return self

    def _difference(self, X: pd.Series, d: int):
        """Helper Function to Difference Time Series n Times"""

        return X if d == 0 else X.diff() if d == 1 else self._difference(X.diff(), d - 1)

    @staticmethod
    def _lagged_values(X: pd.Series, p: int, ar_coef: list):
        """Helper Function to Calculate AutoRegressive(AR) Component"""

        return X if p == 0 else pd.concat([X.copy().shift(periods=i) for i in range(1, p + 1)], axis=1).dot(ar_coef)

    @staticmethod
    def _calculate_residuals(X_ar: pd.Series, X_diff: pd.Series, p: int, d: int, q: int, ma_coef: list):
        """Helper Function to Calculate Residuals/MA Component"""

        ma_coef = ma_coef[::-1]
        resid = X_ar.copy(deep=True)
        resid[:] = 0

        X_ma = X_ar.copy(deep=True)
        X_ma[:] = np.nan

        for x in range(p + d, len(X_ar)):
            ma_component = resid[x - q: x].dot(ma_coef)
            prediction = X_ar[x] + ma_component
            residual = X_diff[x] - prediction
            resid[x] = residual
            X_ma[x] = prediction

        return resid, X_ma

    def _arima_transform_series(self, s: ARIMABestParams) -> pd.Series:
        """Helper Function to Transform Series"""

        # Difference first
        X_diff = self._difference(s.series, s.d)

        # Calculate Autoregressive Component
        X_diff_ar = self._lagged_values(X_diff, s.p, s.ar_coef)

        # Calculate Residuals and Moving Average Component
        _, X_diff_ar_ma = self._calculate_residuals(X_diff_ar, X_diff, s.p, s.d, s.q, s.ma_coef)
        return X_diff_ar_ma

    def transform(self, X: Union[pd.Series, pd.DataFrame]) -> pd.DataFrame:
        """
        Transform a series based on the best ARIMA found from fit().
        Does not support tranformation using MA components.
        :param X: time series to be operated on; required parameter
        :return: DataFrame
        """
        X = X.to_frame() if isinstance(X, pd.Series) else X
        return pd.DataFrame({s_id: self._arima_transform_series(self.best_params[s_id]) for s_id in X.columns})



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
import pandas as pd
from typing import Union

from .core import DataFrameWithInfo, ErrorValue, FloatWithInfo, StringWithInfo, sort_risk
from gs_quant.base import InstrumentBase, RiskKey


_logger = logging.getLogger(__name__)


def __dataframe_handler(field: str, mappings: tuple, result: dict, risk_key: RiskKey) -> DataFrameWithInfo:
    records = [{k: datum[v] for k, v in mappings} for datum in result[field]]
    df = pd.DataFrame.from_records(records)
    return DataFrameWithInfo(sort_risk(df, tuple(k for k, _ in mappings)), risk_key=risk_key)


def cashflows_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase) -> DataFrameWithInfo:
    mappings = (
        ('payment_date', 'payDate'),
        ('set_date', 'setDate'),
        ('accrual_start_date', 'accStart'),
        ('accrual_end_date', 'accEnd'),
        ('payment_amount', 'payAmount'),
        ('notional', 'notional'),
        ('payment_type', 'paymentType'),
        ('floating_rate_option', 'index'),
        ('floating_rate_designated_maturity', 'indexTerm'),
        ('day_count_fraction', 'dayCountFraction'),
        ('spread', 'spread'),
        ('rate', 'rate'),
        ('discount_factor', 'discountFactor')
    )

    for cashflow in result['cashflows']:
        for field in ('payDate', 'setDate', 'accStart', 'accEnd'):
            value = cashflow.get(field)
            date = dt.date.fromisoformat(value) if value else dt.date.max
            cashflow[field] = date

    return __dataframe_handler('cashflows', mappings, result, risk_key)


def error_handler(result: dict, risk_key: RiskKey, instrument: InstrumentBase) -> ErrorValue:
    error = result.get('errorString', 'Unknown error')
    _logger.error(f'Error while computing {risk_key.risk_measure} on {instrument} for {risk_key.date}: {error}')
    return ErrorValue(risk_key, error)


def leg_definition_handler(result: dict, risk_key: RiskKey, instrument: InstrumentBase) -> InstrumentBase:
    new_instrument = instrument.from_dict(result)
    new_instrument.unresolved = instrument
    new_instrument.name = instrument.name
    new_instrument.resolution_key = risk_key
    return new_instrument


def message_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase) -> StringWithInfo:
    return StringWithInfo(risk_key, result.get('message'))


def number_and_unit_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase) -> FloatWithInfo:
    return FloatWithInfo(risk_key, result.get('value', float('nan')), unit=result.get('unit'))


def required_assets_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase):
    mappings = (('mkt_type', 'type'), ('mkt_asset', 'asset'))
    return __dataframe_handler('requiredAssets', mappings, result, risk_key)


def risk_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase) -> FloatWithInfo:
    return FloatWithInfo(risk_key, result.get('val', float('nan')), unit=result.get('unit'))


def risk_by_class_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase)\
        -> Union[DataFrameWithInfo, FloatWithInfo]:
    # TODO Remove this once we migrate parallel USD IRDelta measures
    types = [c['type'] for c in result['classes']]
    if len(types) <= 2 and all(t == 'IR' for t in types):
        return FloatWithInfo(risk_key, sum(result.get('values', (float('nan'),))), unit=result.get('unit'))
    else:
        for clazz, value in zip(result['classes'], result['values']):
            mkt_type = clazz['type']
            if 'SPIKE' in mkt_type or 'JUMP' in mkt_type:
                clazz['type'] = 'OTHER'

            clazz.update({'value': value})

        mappings = (
            ('mkt_type', 'type'),
            ('mkt_asset', 'asset'),
            ('value', 'value')
        )

        return __dataframe_handler('classes', mappings, result, risk_key)


def risk_vector_handler(result: dict, risk_key: RiskKey, _instrument: InstrumentBase) -> DataFrameWithInfo:
    for points, value in zip(result['points'], result['asset']):
        points.update({'value': value})

    mappings = (
        ('mkt_type', 'type'),
        ('mkt_asset', 'asset'),
        ('mkt_class', 'class_'),
        ('mkt_point', 'point'),
        ('value', 'value')
    )

    return __dataframe_handler('points', mappings, result, risk_key)


result_handlers = {
    'Error': error_handler,
    'IRPCashflowTable': cashflows_handler,
    'LegDefinition': leg_definition_handler,
    'Message': message_handler,
    'NumberAndUnit': number_and_unit_handler,
    'RequireAssets': required_assets_handler,
    'Risk': risk_handler,
    'RiskByClass': risk_by_class_handler,
    'RiskVector': risk_vector_handler
}



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gs_quant.target.common import *
import datetime
from typing import Tuple, Union
from enum import Enum
from gs_quant.base import Base, EnumBase, InstrumentBase, camel_case_translate, get_enum_value


class CorporateActionsTypes(EnumBase, Enum):    
    
    """Types of corporate actions in the hedge"""

    Mergers = 'Mergers'
    Quote_lot_adjustments = 'Quote lot adjustments'
    Rights = 'Rights'
    Spinoffs = 'Spinoffs'
    Cash_dividends = 'Cash dividends'
    Stock_splits = 'Stock splits'
    Reorganization = 'Reorganization'
    
    def __repr__(self):
        return self.value


class HedgeObjective(EnumBase, Enum):    
    
    """The objective of the hedge."""

    Minimize_Factor_Risk = 'Minimize Factor Risk'
    Replicate_Performance = 'Replicate Performance'
    
    def __repr__(self):
        return self.value


class HedgeUniverseAssetType(EnumBase, Enum):    
    
    """Type of assets that will be added to the hedge universe."""

    Custom_Basket = 'Custom Basket'
    ETF = 'ETF'
    Research_Basket = 'Research Basket'
    Single_Stock = 'Single Stock'
    
    def __repr__(self):
        return self.value


class HedgerComparisonType(EnumBase, Enum):    
    
    Asset = 'Asset'
    Portfolio = 'Portfolio'
    Hedge = 'Hedge'
    
    def __repr__(self):
        return self.value


class HedgerConstraintPrioritySetting(EnumBase, Enum):    
    
    """Priority of the constraint from 0-5 (prioritized in that order). The
       optimization will fail if it cannot meet a constraint with 0 priority. A
       constraint with priority of 1-5 can be called a relaxed constraint, which
       means that the optimization will make its best effort to meet the
       constraint but will not fail if it cannot."""

    _0 = '0'
    _1 = '1'
    _2 = '2'
    _3 = '3'
    _4 = '4'
    _5 = '5'
    
    def __repr__(self):
        return self.value


class AssetConstraint(Base):
        
    """Constraint on a specific asset in the hedge universe."""

    @camel_case_translate
    def __init__(
        self,
        asset_id: str,
        max_: float,
        min_: float,
        name: str = None
    ):        
        super().__init__()
        self.asset_id = asset_id
        self.__max = max_
        self.__min = min_
        self.name = name

    @property
    def asset_id(self) -> str:
        """Marquee unique asset identifier."""
        return self.__asset_id

    @asset_id.setter
    def asset_id(self, value: str):
        self._property_changed('asset_id')
        self.__asset_id = value        

    @property
    def max(self) -> float:
        """Maximum percentage weight of the asset in the hedge."""
        return self.__max

    @max.setter
    def max(self, value: float):
        self._property_changed('max')
        self.__max = value        

    @property
    def min(self) -> float:
        """Minimum percentage weight of the asset in the hedge."""
        return self.__min

    @min.setter
    def min(self, value: float):
        self._property_changed('min')
        self.__min = value        


class ClassificationConstraint(Base):
        
    """Constraint on an asset classification to be applied to the hedge."""

    @camel_case_translate
    def __init__(
        self,
        type_: str,
        name: str,
        max_: float,
        min_: float
    ):        
        super().__init__()
        self.__type = type_
        self.name = name
        self.__max = max_
        self.__min = min_

    @property
    def type(self) -> str:
        """Type of classification."""
        return self.__type

    @type.setter
    def type(self, value: str):
        self._property_changed('type')
        self.__type = value        

    @property
    def name(self) -> str:
        """Name of classification."""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        

    @property
    def max(self) -> float:
        """Maximum combined weight of any universe assets that fall under this
           classification in the hedge result (0-100)."""
        return self.__max

    @max.setter
    def max(self, value: float):
        self._property_changed('max')
        self.__max = value        

    @property
    def min(self) -> float:
        """Minimum combined weight of any universe assets that fall under this
           classification in the hedge result (0-100)."""
        return self.__min

    @min.setter
    def min(self, value: float):
        self._property_changed('min')
        self.__min = value        


class FactorConstraint(Base):
        
    """Constraint on a specific axioma factor."""

    @camel_case_translate
    def __init__(
        self,
        factor: str,
        exposure: float,
        name: str = None
    ):        
        super().__init__()
        self.factor = factor
        self.exposure = exposure
        self.name = name

    @property
    def factor(self) -> str:
        """Axioma factor name."""
        return self.__factor

    @factor.setter
    def factor(self, value: str):
        self._property_changed('factor')
        self.__factor = value        

    @property
    def exposure(self) -> float:
        """Exposure in USD."""
        return self.__exposure

    @exposure.setter
    def exposure(self, value: float):
        self._property_changed('exposure')
        self.__exposure = value        


class FactorExposure(Base):
        
    """Object representation of a Factor Exposure."""

    @camel_case_translate
    def __init__(
        self,
        factor: str,
        exposure: float,
        name: str = None
    ):        
        super().__init__()
        self.factor = factor
        self.exposure = exposure
        self.name = name

    @property
    def factor(self) -> str:
        """Name of the Axioma factor."""
        return self.__factor

    @factor.setter
    def factor(self, value: str):
        self._property_changed('factor')
        self.__factor = value        

    @property
    def exposure(self) -> float:
        """Exposure in USD."""
        return self.__exposure

    @exposure.setter
    def exposure(self, value: float):
        self._property_changed('exposure')
        self.__exposure = value        


class HedgeBenchmark(Base):
        
    @camel_case_translate
    def __init__(
        self,
        asset_id: str,
        cumulative_pnl: Tuple[Tuple[Union[datetime.date, float], ...], ...] = None,
        name: str = None
    ):        
        super().__init__()
        self.asset_id = asset_id
        self.cumulative_pnl = cumulative_pnl
        self.name = name

    @property
    def asset_id(self) -> str:
        """Marquee unique asset identifier."""
        return self.__asset_id

    @asset_id.setter
    def asset_id(self, value: str):
        self._property_changed('asset_id')
        self.__asset_id = value        

    @property
    def cumulative_pnl(self) -> Tuple[Tuple[Union[datetime.date, float], ...], ...]:
        """An array of tuples that represent values in a time series."""
        return self.__cumulative_pnl

    @cumulative_pnl.setter
    def cumulative_pnl(self, value: Tuple[Tuple[Union[datetime.date, float], ...], ...]):
        self._property_changed('cumulative_pnl')
        self.__cumulative_pnl = value        


class HedgeConstituent(Base):
        
    """Fields returned for each hedge constituent."""

    @camel_case_translate
    def __init__(
        self,
        asset_id: str,
        name: str,
        weight: float,
        currency: Union[Currency, str],
        country: str = None,
        correlation: float = None,
        transaction_cost: float = None,
        marginal_cost: float = None,
        borrow_cost: float = None
    ):        
        super().__init__()
        self.asset_id = asset_id
        self.name = name
        self.weight = weight
        self.currency = currency
        self.country = country
        self.correlation = correlation
        self.transaction_cost = transaction_cost
        self.marginal_cost = marginal_cost
        self.borrow_cost = borrow_cost

    @property
    def asset_id(self) -> str:
        """Marquee unique asset identifier."""
        return self.__asset_id

    @asset_id.setter
    def asset_id(self, value: str):
        self._property_changed('asset_id')
        self.__asset_id = value        

    @property
    def name(self) -> str:
        """Display name of the asset"""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        

    @property
    def weight(self) -> float:
        """Weight of the constituent."""
        return self.__weight

    @weight.setter
    def weight(self, value: float):
        self._property_changed('weight')
        self.__weight = value        

    @property
    def currency(self) -> Union[Currency, str]:
        """Currency, ISO 4217 currency code or exchange quote modifier (e.g. GBP vs GBp)"""
        return self.__currency

    @currency.setter
    def currency(self, value: Union[Currency, str]):
        self._property_changed('currency')
        self.__currency = get_enum_value(Currency, value)        

    @property
    def country(self) -> str:
        """Country name of asset."""
        return self.__country

    @country.setter
    def country(self, value: str):
        self._property_changed('country')
        self.__country = value        

    @property
    def correlation(self) -> float:
        """Max 44 day rolling correlation between the constituent and target returns."""
        return self.__correlation

    @correlation.setter
    def correlation(self, value: float):
        self._property_changed('correlation')
        self.__correlation = value        

    @property
    def transaction_cost(self) -> float:
        """The estimated market impact cost for trading a position or set of positions at a
           specified market participation rate or trade-out period."""
        return self.__transaction_cost

    @transaction_cost.setter
    def transaction_cost(self, value: float):
        self._property_changed('transaction_cost')
        self.__transaction_cost = value        

    @property
    def marginal_cost(self) -> float:
        """The estimated market impact cost multiplied by the position's gross weight in
           the portfolio."""
        return self.__marginal_cost

    @marginal_cost.setter
    def marginal_cost(self, value: float):
        self._property_changed('marginal_cost')
        self.__marginal_cost = value        

    @property
    def borrow_cost(self) -> float:
        """The estimated borrow cost in bps for a position or position set."""
        return self.__borrow_cost

    @borrow_cost.setter
    def borrow_cost(self, value: float):
        self._property_changed('borrow_cost')
        self.__borrow_cost = value        


class HedgerComparisonProperties(Base):
        
    """properties used to hedge the comparison."""

    @camel_case_translate
    def __init__(
        self,
        hedge_value_type: str,
        hedge_value: float,
        name: str = None
    ):        
        super().__init__()
        self.hedge_value_type = hedge_value_type
        self.hedge_value = hedge_value
        self.name = name

    @property
    def hedge_value_type(self) -> str:
        """The type of value for which is applied to value the hedge in the comparison."""
        return self.__hedge_value_type

    @hedge_value_type.setter
    def hedge_value_type(self, value: str):
        self._property_changed('hedge_value_type')
        self.__hedge_value_type = value        

    @property
    def hedge_value(self) -> float:
        return self.__hedge_value

    @hedge_value.setter
    def hedge_value(self, value: float):
        self._property_changed('hedge_value')
        self.__hedge_value = value        


class Target(Base):
        
    """The asset id, portfolio id, or set of positions that make up the hedge target."""

    @camel_case_translate
    def __init__(
        self,
        id_: str = None,
        positions: Tuple[Position, ...] = None,
        name: str = None
    ):        
        super().__init__()
        self.__id = id_
        self.positions = positions
        self.name = name

    @property
    def id(self) -> str:
        """Marquee unique identifier"""
        return self.__id

    @id.setter
    def id(self, value: str):
        self._property_changed('id')
        self.__id = value        

    @property
    def positions(self) -> Tuple[Position, ...]:
        """Array of quantity position objects."""
        return self.__positions

    @positions.setter
    def positions(self, value: Tuple[Position, ...]):
        self._property_changed('positions')
        self.__positions = value        


class FactorExposures(Base):
        
    """Breakdown of factor exposures by category."""

    @camel_case_translate
    def __init__(
        self,
        country: Tuple[FactorExposure, ...],
        industry: Tuple[FactorExposure, ...],
        sector: Tuple[FactorExposure, ...],
        style: Tuple[FactorExposure, ...],
        name: str = None
    ):        
        super().__init__()
        self.country = country
        self.industry = industry
        self.sector = sector
        self.style = style
        self.name = name

    @property
    def country(self) -> Tuple[FactorExposure, ...]:
        """List of country factor exposures."""
        return self.__country

    @country.setter
    def country(self, value: Tuple[FactorExposure, ...]):
        self._property_changed('country')
        self.__country = value        

    @property
    def industry(self) -> Tuple[FactorExposure, ...]:
        """List of industry factor exposures."""
        return self.__industry

    @industry.setter
    def industry(self, value: Tuple[FactorExposure, ...]):
        self._property_changed('industry')
        self.__industry = value        

    @property
    def sector(self) -> Tuple[FactorExposure, ...]:
        """List of sector factor exposures."""
        return self.__sector

    @sector.setter
    def sector(self, value: Tuple[FactorExposure, ...]):
        self._property_changed('sector')
        self.__sector = value        

    @property
    def style(self) -> Tuple[FactorExposure, ...]:
        """List of style factor exposures."""
        return self.__style

    @style.setter
    def style(self, value: Tuple[FactorExposure, ...]):
        self._property_changed('style')
        self.__style = value        


class FactorHedgeUniverse(Base):
        
    """Any combination of asset ids and bulk asset types that make up the universe."""

    @camel_case_translate
    def __init__(
        self,
        asset_ids: Tuple[str, ...] = None,
        asset_types: Tuple[Union[HedgeUniverseAssetType, str], ...] = None,
        name: str = None
    ):        
        super().__init__()
        self.asset_ids = asset_ids
        self.asset_types = asset_types
        self.name = name

    @property
    def asset_ids(self) -> Tuple[str, ...]:
        """Set of marquee unique asset identifiers that make up the hedge universe. Any
           identifiers that are not for single stocks will be exploded into
           their underliers."""
        return self.__asset_ids

    @asset_ids.setter
    def asset_ids(self, value: Tuple[str, ...]):
        self._property_changed('asset_ids')
        self.__asset_ids = value        

    @property
    def asset_types(self) -> Tuple[Union[HedgeUniverseAssetType, str], ...]:
        """Types of assets that will be added to the hedge universe."""
        return self.__asset_types

    @asset_types.setter
    def asset_types(self, value: Tuple[Union[HedgeUniverseAssetType, str], ...]):
        self._property_changed('asset_types')
        self.__asset_types = value        


class FactorHedgerConstraintPrioritySettings(Base):
        
    """Specify the priority of constraints"""

    @camel_case_translate
    def __init__(
        self,
        min_sector_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        max_sector_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        min_industry_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        max_industry_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        min_region_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        max_region_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        min_country_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        max_country_weights: Union[HedgerConstraintPrioritySetting, str] = None,
        style_exposures: Union[HedgerConstraintPrioritySetting, str] = None,
        country_exposures: Union[HedgerConstraintPrioritySetting, str] = None,
        region_exposures: Union[HedgerConstraintPrioritySetting, str] = None,
        industry_exposures: Union[HedgerConstraintPrioritySetting, str] = None,
        sector_exposures: Union[HedgerConstraintPrioritySetting, str] = None,
        name: str = None
    ):        
        super().__init__()
        self.min_sector_weights = min_sector_weights
        self.max_sector_weights = max_sector_weights
        self.min_industry_weights = min_industry_weights
        self.max_industry_weights = max_industry_weights
        self.min_region_weights = min_region_weights
        self.max_region_weights = max_region_weights
        self.min_country_weights = min_country_weights
        self.max_country_weights = max_country_weights
        self.style_exposures = style_exposures
        self.country_exposures = country_exposures
        self.region_exposures = region_exposures
        self.industry_exposures = industry_exposures
        self.sector_exposures = sector_exposures
        self.name = name

    @property
    def min_sector_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__min_sector_weights

    @min_sector_weights.setter
    def min_sector_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('min_sector_weights')
        self.__min_sector_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def max_sector_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__max_sector_weights

    @max_sector_weights.setter
    def max_sector_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('max_sector_weights')
        self.__max_sector_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def min_industry_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__min_industry_weights

    @min_industry_weights.setter
    def min_industry_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('min_industry_weights')
        self.__min_industry_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def max_industry_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__max_industry_weights

    @max_industry_weights.setter
    def max_industry_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('max_industry_weights')
        self.__max_industry_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def min_region_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__min_region_weights

    @min_region_weights.setter
    def min_region_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('min_region_weights')
        self.__min_region_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def max_region_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__max_region_weights

    @max_region_weights.setter
    def max_region_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('max_region_weights')
        self.__max_region_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def min_country_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__min_country_weights

    @min_country_weights.setter
    def min_country_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('min_country_weights')
        self.__min_country_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def max_country_weights(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__max_country_weights

    @max_country_weights.setter
    def max_country_weights(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('max_country_weights')
        self.__max_country_weights = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def style_exposures(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__style_exposures

    @style_exposures.setter
    def style_exposures(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('style_exposures')
        self.__style_exposures = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def country_exposures(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__country_exposures

    @country_exposures.setter
    def country_exposures(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('country_exposures')
        self.__country_exposures = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def region_exposures(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__region_exposures

    @region_exposures.setter
    def region_exposures(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('region_exposures')
        self.__region_exposures = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def industry_exposures(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__industry_exposures

    @industry_exposures.setter
    def industry_exposures(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('industry_exposures')
        self.__industry_exposures = get_enum_value(HedgerConstraintPrioritySetting, value)        

    @property
    def sector_exposures(self) -> Union[HedgerConstraintPrioritySetting, str]:
        """Priority of the constraint from 0-5 (prioritized in that order). The
           optimization will fail if it cannot meet a constraint with 0
           priority. A constraint with priority of 1-5 can be called a relaxed
           constraint, which means that the optimization will make its best
           effort to meet the constraint but will not fail if it cannot."""
        return self.__sector_exposures

    @sector_exposures.setter
    def sector_exposures(self, value: Union[HedgerConstraintPrioritySetting, str]):
        self._property_changed('sector_exposures')
        self.__sector_exposures = get_enum_value(HedgerConstraintPrioritySetting, value)        


class PerformanceHedgeResult(Base):
        
    """Result of a performance replication hedge."""

    @camel_case_translate
    def __init__(
        self,
        target: dict,
        hedge: dict = None,
        hedged_target: dict = None,
        benchmarks: Tuple[dict, ...] = None,
        name: str = None
    ):        
        super().__init__()
        self.target = target
        self.hedge = hedge
        self.hedged_target = hedged_target
        self.benchmarks = benchmarks
        self.name = name

    @property
    def target(self) -> dict:
        """Statistics of the target."""
        return self.__target

    @target.setter
    def target(self, value: dict):
        self._property_changed('target')
        self.__target = value        

    @property
    def hedge(self) -> dict:
        """Statistics of the portfolio of assets making up the hedge."""
        return self.__hedge

    @hedge.setter
    def hedge(self, value: dict):
        self._property_changed('hedge')
        self.__hedge = value        

    @property
    def hedged_target(self) -> dict:
        """Statistics of the portfolio resulting from combining the target and hedge
           portfolios."""
        return self.__hedged_target

    @hedged_target.setter
    def hedged_target(self, value: dict):
        self._property_changed('hedged_target')
        self.__hedged_target = value        

    @property
    def benchmarks(self) -> Tuple[dict, ...]:
        """Benchmarks compared against the target."""
        return self.__benchmarks

    @benchmarks.setter
    def benchmarks(self, value: Tuple[dict, ...]):
        self._property_changed('benchmarks')
        self.__benchmarks = value        


class FactorHedgerResultPositions(Base):
        
    """The set of values returned for either the target, hedge, or hedged target
       positions in a hedge calculation with an objective to minimize factor
       risk."""

    @camel_case_translate
    def __init__(
        self,
        beta_exposure: float,
        daily_va_r: float,
        factor_exposures: FactorExposures,
        specific_exposure: float,
        systematic_exposure: float,
        total_risk: float,
        volatility: float,
        net_exposure: float,
        constituents: Tuple[HedgeConstituent, ...] = None,
        number_of_positions: float = None,
        cumulative_pnl: Tuple[Tuple[Union[datetime.date, float], ...], ...] = None,
        transaction_cost: float = None,
        borrow_cost_bps: float = None,
        max_drawdown: float = None,
        gross_exposure: float = None,
        long_exposure: float = None,
        short_exposure: float = None,
        tracking_error: float = None,
        correlation: float = None,
        exposure_overlap_with_target: float = None,
        total_pnl: float = None,
        name: str = None
    ):        
        super().__init__()
        self.beta_exposure = beta_exposure
        self.constituents = constituents
        self.number_of_positions = number_of_positions
        self.daily_va_r = daily_va_r
        self.factor_exposures = factor_exposures
        self.cumulative_pnl = cumulative_pnl
        self.specific_exposure = specific_exposure
        self.systematic_exposure = systematic_exposure
        self.total_risk = total_risk
        self.transaction_cost = transaction_cost
        self.borrow_cost_bps = borrow_cost_bps
        self.volatility = volatility
        self.max_drawdown = max_drawdown
        self.net_exposure = net_exposure
        self.gross_exposure = gross_exposure
        self.long_exposure = long_exposure
        self.short_exposure = short_exposure
        self.tracking_error = tracking_error
        self.correlation = correlation
        self.exposure_overlap_with_target = exposure_overlap_with_target
        self.total_pnl = total_pnl
        self.name = name

    @property
    def beta_exposure(self) -> float:
        """Exposure to beta."""
        return self.__beta_exposure

    @beta_exposure.setter
    def beta_exposure(self, value: float):
        self._property_changed('beta_exposure')
        self.__beta_exposure = value        

    @property
    def constituents(self) -> Tuple[HedgeConstituent, ...]:
        """Composition of asset positions that make up the hedge."""
        return self.__constituents

    @constituents.setter
    def constituents(self, value: Tuple[HedgeConstituent, ...]):
        self._property_changed('constituents')
        self.__constituents = value        

    @property
    def number_of_positions(self) -> float:
        """Number of positions in the entity."""
        return self.__number_of_positions

    @number_of_positions.setter
    def number_of_positions(self, value: float):
        self._property_changed('number_of_positions')
        self.__number_of_positions = value        

    @property
    def daily_va_r(self) -> float:
        """Daily value at risk."""
        return self.__daily_va_r

    @daily_va_r.setter
    def daily_va_r(self, value: float):
        self._property_changed('daily_va_r')
        self.__daily_va_r = value        

    @property
    def factor_exposures(self) -> FactorExposures:
        """Breakdown of factor exposures by category."""
        return self.__factor_exposures

    @factor_exposures.setter
    def factor_exposures(self, value: FactorExposures):
        self._property_changed('factor_exposures')
        self.__factor_exposures = value        

    @property
    def cumulative_pnl(self) -> Tuple[Tuple[Union[datetime.date, float], ...], ...]:
        """Time series of cumulative Pnl."""
        return self.__cumulative_pnl

    @cumulative_pnl.setter
    def cumulative_pnl(self, value: Tuple[Tuple[Union[datetime.date, float], ...], ...]):
        self._property_changed('cumulative_pnl')
        self.__cumulative_pnl = value        

    @property
    def specific_exposure(self) -> float:
        """Change in portfolio value which is idiosyncratic to individual assets."""
        return self.__specific_exposure

    @specific_exposure.setter
    def specific_exposure(self, value: float):
        self._property_changed('specific_exposure')
        self.__specific_exposure = value        

    @property
    def systematic_exposure(self) -> float:
        """Change in portfolio value attributable to systematic risk factors."""
        return self.__systematic_exposure

    @systematic_exposure.setter
    def systematic_exposure(self, value: float):
        self._property_changed('systematic_exposure')
        self.__systematic_exposure = value        

    @property
    def total_risk(self) -> float:
        """Cumulative daily fluctuation in portfolio value."""
        return self.__total_risk

    @total_risk.setter
    def total_risk(self, value: float):
        self._property_changed('total_risk')
        self.__total_risk = value        

    @property
    def transaction_cost(self) -> float:
        """The estimated market impact cost for trading a position or set of positions at a
           specified market participation rate or trade-out period."""
        return self.__transaction_cost

    @transaction_cost.setter
    def transaction_cost(self, value: float):
        self._property_changed('transaction_cost')
        self.__transaction_cost = value        

    @property
    def borrow_cost_bps(self) -> float:
        """The estimated borrow cost in bps for a position or position set."""
        return self.__borrow_cost_bps

    @borrow_cost_bps.setter
    def borrow_cost_bps(self, value: float):
        self._property_changed('borrow_cost_bps')
        self.__borrow_cost_bps = value        

    @property
    def volatility(self) -> float:
        """Standard deviation of the annualized returns."""
        return self.__volatility

    @volatility.setter
    def volatility(self, value: float):
        self._property_changed('volatility')
        self.__volatility = value        

    @property
    def max_drawdown(self) -> float:
        """Maximum loss from a peak to a trough within the specified dates."""
        return self.__max_drawdown

    @max_drawdown.setter
    def max_drawdown(self, value: float):
        self._property_changed('max_drawdown')
        self.__max_drawdown = value        

    @property
    def net_exposure(self) -> float:
        """Combined exposure of long and short positions."""
        return self.__net_exposure

    @net_exposure.setter
    def net_exposure(self, value: float):
        self._property_changed('net_exposure')
        self.__net_exposure = value        

    @property
    def gross_exposure(self) -> float:
        """Combined exposure of long and short positions."""
        return self.__gross_exposure

    @gross_exposure.setter
    def gross_exposure(self, value: float):
        self._property_changed('gross_exposure')
        self.__gross_exposure = value        

    @property
    def long_exposure(self) -> float:
        """Long exposure of the position set."""
        return self.__long_exposure

    @long_exposure.setter
    def long_exposure(self, value: float):
        self._property_changed('long_exposure')
        self.__long_exposure = value        

    @property
    def short_exposure(self) -> float:
        """Short exposure of the position set."""
        return self.__short_exposure

    @short_exposure.setter
    def short_exposure(self, value: float):
        self._property_changed('short_exposure')
        self.__short_exposure = value        

    @property
    def tracking_error(self) -> float:
        """Standard deviation of the difference in the portfolio and benchmark returns over
           time."""
        return self.__tracking_error

    @tracking_error.setter
    def tracking_error(self, value: float):
        self._property_changed('tracking_error')
        self.__tracking_error = value        

    @property
    def correlation(self) -> float:
        """Max 44 day rolling correlation."""
        return self.__correlation

    @correlation.setter
    def correlation(self, value: float):
        self._property_changed('correlation')
        self.__correlation = value        

    @property
    def exposure_overlap_with_target(self) -> float:
        """Overlapping exposure with the target portfolio."""
        return self.__exposure_overlap_with_target

    @exposure_overlap_with_target.setter
    def exposure_overlap_with_target(self, value: float):
        self._property_changed('exposure_overlap_with_target')
        self.__exposure_overlap_with_target = value        

    @property
    def total_pnl(self) -> float:
        """Total PnL of the position set over the backtest period."""
        return self.__total_pnl

    @total_pnl.setter
    def total_pnl(self, value: float):
        self._property_changed('total_pnl')
        self.__total_pnl = value        


class FactorHedgeResult(Base):
        
    """Result of a hedge calculation with an objective to minimize factor risk."""

    @camel_case_translate
    def __init__(
        self,
        hedge: FactorHedgerResultPositions,
        hedged_target: FactorHedgerResultPositions,
        target: FactorHedgerResultPositions = None,
        name: str = None
    ):        
        super().__init__()
        self.target = target
        self.hedge = hedge
        self.hedged_target = hedged_target
        self.name = name

    @property
    def target(self) -> FactorHedgerResultPositions:
        """Statistics of the target portfolio."""
        return self.__target

    @target.setter
    def target(self, value: FactorHedgerResultPositions):
        self._property_changed('target')
        self.__target = value        

    @property
    def hedge(self) -> FactorHedgerResultPositions:
        """Statistics of the portfolio of assets making up the hedge."""
        return self.__hedge

    @hedge.setter
    def hedge(self, value: FactorHedgerResultPositions):
        self._property_changed('hedge')
        self.__hedge = value        

    @property
    def hedged_target(self) -> FactorHedgerResultPositions:
        """Statistics of the portfolio resulting from combining the target and hedge
           portfolios."""
        return self.__hedged_target

    @hedged_target.setter
    def hedged_target(self, value: FactorHedgerResultPositions):
        self._property_changed('hedged_target')
        self.__hedged_target = value        


class HedgerComparison(Base):
        
    """Comparison of a hedge."""

    @camel_case_translate
    def __init__(
        self,
        entity_id: str,
        entity_type: Union[HedgerComparisonType, str],
        hedge_properties: HedgerComparisonProperties,
        result: dict = None,
        name: str = None
    ):        
        super().__init__()
        self.entity_id = entity_id
        self.entity_type = entity_type
        self.hedge_properties = hedge_properties
        self.result = result
        self.name = name

    @property
    def entity_id(self) -> str:
        """Id of the marquee entity"""
        return self.__entity_id

    @entity_id.setter
    def entity_id(self, value: str):
        self._property_changed('entity_id')
        self.__entity_id = value        

    @property
    def entity_type(self) -> Union[HedgerComparisonType, str]:
        return self.__entity_type

    @entity_type.setter
    def entity_type(self, value: Union[HedgerComparisonType, str]):
        self._property_changed('entity_type')
        self.__entity_type = get_enum_value(HedgerComparisonType, value)        

    @property
    def hedge_properties(self) -> HedgerComparisonProperties:
        """properties used to hedge the comparison."""
        return self.__hedge_properties

    @hedge_properties.setter
    def hedge_properties(self, value: HedgerComparisonProperties):
        self._property_changed('hedge_properties')
        self.__hedge_properties = value        

    @property
    def result(self) -> dict:
        """Result of the hedge comparison."""
        return self.__result

    @result.setter
    def result(self, value: dict):
        self._property_changed('result')
        self.__result = value        


class FactorHedgeParameters(Base):
        
    """Parameters for a hedge calculation to minimize factor risk."""

    @camel_case_translate
    def __init__(
        self,
        risk_model: str,
        target_notional: float,
        hedge_notional: float,
        hedge_target: Target,
        hedge_universe: FactorHedgeUniverse,
        hedge_date: datetime.date,
        backtest_start_date: datetime.date,
        backtest_end_date: datetime.date,
        fx_hedged: bool = None,
        exclude_target_assets: bool = None,
        exclude_corporate_actions: bool = None,
        exclude_corporate_actions_types: Tuple[Union[CorporateActionsTypes, str], ...] = None,
        exclude_hard_to_borrow_assets: bool = None,
        exclude_restricted_assets: bool = None,
        max_adv_percentage: float = None,
        explode_universe: bool = None,
        min_names: float = None,
        max_names: float = None,
        min_weight: float = None,
        max_weight: float = None,
        min_market_cap: float = None,
        max_market_cap: float = None,
        market_participation_rate: float = 10,
        asset_constraints: Tuple[AssetConstraint, ...] = None,
        factor_constraints: Tuple[FactorConstraint, ...] = None,
        classification_constraints: Tuple[ClassificationConstraint, ...] = None,
        constraint_priority_settings: FactorHedgerConstraintPrioritySettings = None,
        comparisons: Tuple[HedgerComparison, ...] = None,
        name: str = None
    ):        
        super().__init__()
        self.risk_model = risk_model
        self.hedge_target = hedge_target
        self.target_notional = target_notional
        self.hedge_notional = hedge_notional
        self.hedge_universe = hedge_universe
        self.hedge_date = hedge_date
        self.backtest_start_date = backtest_start_date
        self.backtest_end_date = backtest_end_date
        self.fx_hedged = fx_hedged
        self.exclude_target_assets = exclude_target_assets
        self.exclude_corporate_actions = exclude_corporate_actions
        self.exclude_corporate_actions_types = exclude_corporate_actions_types
        self.exclude_hard_to_borrow_assets = exclude_hard_to_borrow_assets
        self.exclude_restricted_assets = exclude_restricted_assets
        self.max_adv_percentage = max_adv_percentage
        self.explode_universe = explode_universe
        self.min_names = min_names
        self.max_names = max_names
        self.min_weight = min_weight
        self.max_weight = max_weight
        self.min_market_cap = min_market_cap
        self.max_market_cap = max_market_cap
        self.market_participation_rate = market_participation_rate
        self.asset_constraints = asset_constraints
        self.factor_constraints = factor_constraints
        self.classification_constraints = classification_constraints
        self.constraint_priority_settings = constraint_priority_settings
        self.comparisons = comparisons
        self.name = name

    @property
    def risk_model(self) -> str:
        """Marquee unique risk model identifier"""
        return self.__risk_model

    @risk_model.setter
    def risk_model(self, value: str):
        self._property_changed('risk_model')
        self.__risk_model = value        

    @property
    def hedge_target(self) -> Target:
        """The asset id, portfolio id, or set of positions that make up the hedge target."""
        return self.__hedge_target

    @hedge_target.setter
    def hedge_target(self, value: Target):
        self._property_changed('hedge_target')
        self.__hedge_target = value        

    @property
    def target_notional(self) -> float:
        """Notional value of the hedge target."""
        return self.__target_notional

    @target_notional.setter
    def target_notional(self, value: float):
        self._property_changed('target_notional')
        self.__target_notional = value        

    @property
    def hedge_notional(self) -> float:
        """Notional value of the hedge."""
        return self.__hedge_notional

    @hedge_notional.setter
    def hedge_notional(self, value: float):
        self._property_changed('hedge_notional')
        self.__hedge_notional = value        

    @property
    def hedge_universe(self) -> FactorHedgeUniverse:
        """Any combination of asset ids and bulk asset types that make up the universe."""
        return self.__hedge_universe

    @hedge_universe.setter
    def hedge_universe(self, value: FactorHedgeUniverse):
        self._property_changed('hedge_universe')
        self.__hedge_universe = value        

    @property
    def hedge_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__hedge_date

    @hedge_date.setter
    def hedge_date(self, value: datetime.date):
        self._property_changed('hedge_date')
        self.__hedge_date = value        

    @property
    def backtest_start_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__backtest_start_date

    @backtest_start_date.setter
    def backtest_start_date(self, value: datetime.date):
        self._property_changed('backtest_start_date')
        self.__backtest_start_date = value        

    @property
    def backtest_end_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__backtest_end_date

    @backtest_end_date.setter
    def backtest_end_date(self, value: datetime.date):
        self._property_changed('backtest_end_date')
        self.__backtest_end_date = value        

    @property
    def fx_hedged(self) -> bool:
        """Assume portfolio is FX Hedged."""
        return self.__fx_hedged

    @fx_hedged.setter
    def fx_hedged(self, value: bool):
        self._property_changed('fx_hedged')
        self.__fx_hedged = value        

    @property
    def exclude_target_assets(self) -> bool:
        """Exclude assets in the target composition from being in the hedge."""
        return self.__exclude_target_assets

    @exclude_target_assets.setter
    def exclude_target_assets(self, value: bool):
        self._property_changed('exclude_target_assets')
        self.__exclude_target_assets = value        

    @property
    def exclude_corporate_actions(self) -> bool:
        """Exclude assets pending for corporate actions from being in the hedge"""
        return self.__exclude_corporate_actions

    @exclude_corporate_actions.setter
    def exclude_corporate_actions(self, value: bool):
        self._property_changed('exclude_corporate_actions')
        self.__exclude_corporate_actions = value        

    @property
    def exclude_corporate_actions_types(self) -> Tuple[Union[CorporateActionsTypes, str], ...]:
        """Set of of corporate actions to be excluded in the hedge"""
        return self.__exclude_corporate_actions_types

    @exclude_corporate_actions_types.setter
    def exclude_corporate_actions_types(self, value: Tuple[Union[CorporateActionsTypes, str], ...]):
        self._property_changed('exclude_corporate_actions_types')
        self.__exclude_corporate_actions_types = value        

    @property
    def exclude_hard_to_borrow_assets(self) -> bool:
        """Whether hard to borrow assets should be excluded in the universe or not. True
           for exclude."""
        return self.__exclude_hard_to_borrow_assets

    @exclude_hard_to_borrow_assets.setter
    def exclude_hard_to_borrow_assets(self, value: bool):
        self._property_changed('exclude_hard_to_borrow_assets')
        self.__exclude_hard_to_borrow_assets = value        

    @property
    def exclude_restricted_assets(self) -> bool:
        """Whether to include assets in restricted trading lists or not."""
        return self.__exclude_restricted_assets

    @exclude_restricted_assets.setter
    def exclude_restricted_assets(self, value: bool):
        self._property_changed('exclude_restricted_assets')
        self.__exclude_restricted_assets = value        

    @property
    def max_adv_percentage(self) -> float:
        """Maximum percentage notional to average daily dollar volume allowed for any hedge
           constituent."""
        return self.__max_adv_percentage

    @max_adv_percentage.setter
    def max_adv_percentage(self, value: float):
        self._property_changed('max_adv_percentage')
        self.__max_adv_percentage = value        

    @property
    def explode_universe(self) -> bool:
        """Explode the assets in the universe into their underliers to be used as the hedge
           universe."""
        return self.__explode_universe

    @explode_universe.setter
    def explode_universe(self, value: bool):
        self._property_changed('explode_universe')
        self.__explode_universe = value        

    @property
    def min_names(self) -> float:
        """Minimum number of assets allowed in the hedge."""
        return self.__min_names

    @min_names.setter
    def min_names(self, value: float):
        self._property_changed('min_names')
        self.__min_names = value        

    @property
    def max_names(self) -> float:
        """Maximum number of assets allowed in the hedge."""
        return self.__max_names

    @max_names.setter
    def max_names(self, value: float):
        self._property_changed('max_names')
        self.__max_names = value        

    @property
    def min_weight(self) -> float:
        """Minimum weight of any constituent in hedge."""
        return self.__min_weight

    @min_weight.setter
    def min_weight(self, value: float):
        self._property_changed('min_weight')
        self.__min_weight = value        

    @property
    def max_weight(self) -> float:
        """Maximum weight of any constituent in hedge."""
        return self.__max_weight

    @max_weight.setter
    def max_weight(self, value: float):
        self._property_changed('max_weight')
        self.__max_weight = value        

    @property
    def min_market_cap(self) -> float:
        """Lowest market cap allowed for any hedge constituent."""
        return self.__min_market_cap

    @min_market_cap.setter
    def min_market_cap(self, value: float):
        self._property_changed('min_market_cap')
        self.__min_market_cap = value        

    @property
    def max_market_cap(self) -> float:
        """Highest market cap allowed for any hedge constituent."""
        return self.__max_market_cap

    @max_market_cap.setter
    def max_market_cap(self, value: float):
        self._property_changed('max_market_cap')
        self.__max_market_cap = value        

    @property
    def market_participation_rate(self) -> float:
        """Maximum market participation rate used to estimate the cost of trading a
           portfolio of stocks."""
        return self.__market_participation_rate

    @market_participation_rate.setter
    def market_participation_rate(self, value: float):
        self._property_changed('market_participation_rate')
        self.__market_participation_rate = value        

    @property
    def asset_constraints(self) -> Tuple[AssetConstraint, ...]:
        """Constraints to be applied to specific assets in the hedge universe."""
        return self.__asset_constraints

    @asset_constraints.setter
    def asset_constraints(self, value: Tuple[AssetConstraint, ...]):
        self._property_changed('asset_constraints')
        self.__asset_constraints = value        

    @property
    def factor_constraints(self) -> Tuple[FactorConstraint, ...]:
        """Constraints to be applied to specific axioma factors in the hedge."""
        return self.__factor_constraints

    @factor_constraints.setter
    def factor_constraints(self, value: Tuple[FactorConstraint, ...]):
        self._property_changed('factor_constraints')
        self.__factor_constraints = value        

    @property
    def classification_constraints(self) -> Tuple[ClassificationConstraint, ...]:
        """Constraints to be applied to assets in the universe that are classified by the
           industry, sector, country, or region in the constraint."""
        return self.__classification_constraints

    @classification_constraints.setter
    def classification_constraints(self, value: Tuple[ClassificationConstraint, ...]):
        self._property_changed('classification_constraints')
        self.__classification_constraints = value        

    @property
    def constraint_priority_settings(self) -> FactorHedgerConstraintPrioritySettings:
        """Specify the priority of constraints"""
        return self.__constraint_priority_settings

    @constraint_priority_settings.setter
    def constraint_priority_settings(self, value: FactorHedgerConstraintPrioritySettings):
        self._property_changed('constraint_priority_settings')
        self.__constraint_priority_settings = value        

    @property
    def comparisons(self) -> Tuple[HedgerComparison, ...]:
        """Specifies the list of assets, hedges, or portfolios to compare the hedge
           against."""
        return self.__comparisons

    @comparisons.setter
    def comparisons(self, value: Tuple[HedgerComparison, ...]):
        self._property_changed('comparisons')
        self.__comparisons = value        


class PerformanceHedgeParameters(Base):
        
    """Parameters for a performance replication hedge calculation."""

    @camel_case_translate
    def __init__(
        self,
        hedge_target: Target,
        universe: Tuple[str, ...],
        notional: float,
        observation_start_date: datetime.date,
        observation_end_date: datetime.date,
        max_leverage: float = None,
        backtest_start_date: datetime.date = None,
        backtest_end_date: datetime.date = None,
        sampling_period: str = 'Weekly',
        exclude_target_assets: bool = None,
        exclude_corporate_actions: bool = None,
        exclude_corporate_actions_types: Tuple[Union[CorporateActionsTypes, str], ...] = None,
        exclude_hard_to_borrow_assets: bool = None,
        exclude_restricted_assets: bool = None,
        max_adv_percentage: float = None,
        explode_universe: bool = None,
        max_return_deviation: float = None,
        max_weight: float = None,
        min_market_cap: float = None,
        max_market_cap: float = None,
        market_participation_rate: float = 10,
        asset_constraints: Tuple[AssetConstraint, ...] = None,
        classification_constraints: Tuple[ClassificationConstraint, ...] = None,
        benchmarks: Tuple[str, ...] = None,
        use_machine_learning: bool = False,
        lasso_weight: float = None,
        ridge_weight: float = None,
        name: str = None
    ):        
        super().__init__()
        self.hedge_target = hedge_target
        self.universe = universe
        self.notional = notional
        self.max_leverage = max_leverage
        self.observation_start_date = observation_start_date
        self.observation_end_date = observation_end_date
        self.backtest_start_date = backtest_start_date
        self.backtest_end_date = backtest_end_date
        self.sampling_period = sampling_period
        self.exclude_target_assets = exclude_target_assets
        self.exclude_corporate_actions = exclude_corporate_actions
        self.exclude_corporate_actions_types = exclude_corporate_actions_types
        self.exclude_hard_to_borrow_assets = exclude_hard_to_borrow_assets
        self.exclude_restricted_assets = exclude_restricted_assets
        self.max_adv_percentage = max_adv_percentage
        self.explode_universe = explode_universe
        self.max_return_deviation = max_return_deviation
        self.max_weight = max_weight
        self.min_market_cap = min_market_cap
        self.max_market_cap = max_market_cap
        self.market_participation_rate = market_participation_rate
        self.asset_constraints = asset_constraints
        self.classification_constraints = classification_constraints
        self.benchmarks = benchmarks
        self.use_machine_learning = use_machine_learning
        self.lasso_weight = lasso_weight
        self.ridge_weight = ridge_weight
        self.name = name

    @property
    def hedge_target(self) -> Target:
        """The asset id, portfolio id, or set of positions that make up the hedge target."""
        return self.__hedge_target

    @hedge_target.setter
    def hedge_target(self, value: Target):
        self._property_changed('hedge_target')
        self.__hedge_target = value        

    @property
    def universe(self) -> Tuple[str, ...]:
        """Set of marquee unique asset identifiers that make up the hedge universe. Any
           identifiers that are not for single stocks will be exploded into
           their underliers."""
        return self.__universe

    @universe.setter
    def universe(self, value: Tuple[str, ...]):
        self._property_changed('universe')
        self.__universe = value        

    @property
    def notional(self) -> float:
        """Notional value of the hedge target."""
        return self.__notional

    @notional.setter
    def notional(self, value: float):
        self._property_changed('notional')
        self.__notional = value        

    @property
    def max_leverage(self) -> float:
        """Maximum percentage of the notional that can be used to hedge."""
        return self.__max_leverage

    @max_leverage.setter
    def max_leverage(self, value: float):
        self._property_changed('max_leverage')
        self.__max_leverage = value        

    @property
    def observation_start_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__observation_start_date

    @observation_start_date.setter
    def observation_start_date(self, value: datetime.date):
        self._property_changed('observation_start_date')
        self.__observation_start_date = value        

    @property
    def observation_end_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__observation_end_date

    @observation_end_date.setter
    def observation_end_date(self, value: datetime.date):
        self._property_changed('observation_end_date')
        self.__observation_end_date = value        

    @property
    def backtest_start_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__backtest_start_date

    @backtest_start_date.setter
    def backtest_start_date(self, value: datetime.date):
        self._property_changed('backtest_start_date')
        self.__backtest_start_date = value        

    @property
    def backtest_end_date(self) -> datetime.date:
        """ISO 8601-formatted date"""
        return self.__backtest_end_date

    @backtest_end_date.setter
    def backtest_end_date(self, value: datetime.date):
        self._property_changed('backtest_end_date')
        self.__backtest_end_date = value        

    @property
    def sampling_period(self) -> str:
        """The length of time in between return samples."""
        return self.__sampling_period

    @sampling_period.setter
    def sampling_period(self, value: str):
        self._property_changed('sampling_period')
        self.__sampling_period = value        

    @property
    def exclude_target_assets(self) -> bool:
        """Exclude assets in the target composition from being in the hedge."""
        return self.__exclude_target_assets

    @exclude_target_assets.setter
    def exclude_target_assets(self, value: bool):
        self._property_changed('exclude_target_assets')
        self.__exclude_target_assets = value        

    @property
    def exclude_corporate_actions(self) -> bool:
        """Exclude assets pending for corporate actions from being in the hedge"""
        return self.__exclude_corporate_actions

    @exclude_corporate_actions.setter
    def exclude_corporate_actions(self, value: bool):
        self._property_changed('exclude_corporate_actions')
        self.__exclude_corporate_actions = value        

    @property
    def exclude_corporate_actions_types(self) -> Tuple[Union[CorporateActionsTypes, str], ...]:
        """Set of of corporate actions to be excluded in the hedge"""
        return self.__exclude_corporate_actions_types

    @exclude_corporate_actions_types.setter
    def exclude_corporate_actions_types(self, value: Tuple[Union[CorporateActionsTypes, str], ...]):
        self._property_changed('exclude_corporate_actions_types')
        self.__exclude_corporate_actions_types = value        

    @property
    def exclude_hard_to_borrow_assets(self) -> bool:
        """Whether hard to borrow assets should be excluded in the universe or not. True
           for exclude."""
        return self.__exclude_hard_to_borrow_assets

    @exclude_hard_to_borrow_assets.setter
    def exclude_hard_to_borrow_assets(self, value: bool):
        self._property_changed('exclude_hard_to_borrow_assets')
        self.__exclude_hard_to_borrow_assets = value        

    @property
    def exclude_restricted_assets(self) -> bool:
        """Whether to include assets in restricted trading lists or not."""
        return self.__exclude_restricted_assets

    @exclude_restricted_assets.setter
    def exclude_restricted_assets(self, value: bool):
        self._property_changed('exclude_restricted_assets')
        self.__exclude_restricted_assets = value        

    @property
    def max_adv_percentage(self) -> float:
        """Maximum percentage notional to average daily dollar volume allowed for any hedge
           constituent."""
        return self.__max_adv_percentage

    @max_adv_percentage.setter
    def max_adv_percentage(self, value: float):
        self._property_changed('max_adv_percentage')
        self.__max_adv_percentage = value        

    @property
    def explode_universe(self) -> bool:
        """Explode the assets in the universe into their underliers to be used as the hedge
           universe."""
        return self.__explode_universe

    @explode_universe.setter
    def explode_universe(self, value: bool):
        self._property_changed('explode_universe')
        self.__explode_universe = value        

    @property
    def max_return_deviation(self) -> float:
        """Maximum percentage difference in annualized return between the target and the
           hedge result."""
        return self.__max_return_deviation

    @max_return_deviation.setter
    def max_return_deviation(self, value: float):
        self._property_changed('max_return_deviation')
        self.__max_return_deviation = value        

    @property
    def max_weight(self) -> float:
        """Maximum weight of any constituent in hedge."""
        return self.__max_weight

    @max_weight.setter
    def max_weight(self, value: float):
        self._property_changed('max_weight')
        self.__max_weight = value        

    @property
    def min_market_cap(self) -> float:
        """Lowest market cap allowed for any hedge constituent."""
        return self.__min_market_cap

    @min_market_cap.setter
    def min_market_cap(self, value: float):
        self._property_changed('min_market_cap')
        self.__min_market_cap = value        

    @property
    def max_market_cap(self) -> float:
        """Highest market cap allowed for any hedge constituent."""
        return self.__max_market_cap

    @max_market_cap.setter
    def max_market_cap(self, value: float):
        self._property_changed('max_market_cap')
        self.__max_market_cap = value        

    @property
    def market_participation_rate(self) -> float:
        """Maximum market participation rate used to estimate the cost of trading a
           portfolio of stocks."""
        return self.__market_participation_rate

    @market_participation_rate.setter
    def market_participation_rate(self, value: float):
        self._property_changed('market_participation_rate')
        self.__market_participation_rate = value        

    @property
    def asset_constraints(self) -> Tuple[AssetConstraint, ...]:
        """Constraints to be applied to specific assets in the hedge universe."""
        return self.__asset_constraints

    @asset_constraints.setter
    def asset_constraints(self, value: Tuple[AssetConstraint, ...]):
        self._property_changed('asset_constraints')
        self.__asset_constraints = value        

    @property
    def classification_constraints(self) -> Tuple[ClassificationConstraint, ...]:
        """Constraints to be applied to assets in the universe that are classified by the
           industry, sector, country, or region in the constraint."""
        return self.__classification_constraints

    @classification_constraints.setter
    def classification_constraints(self, value: Tuple[ClassificationConstraint, ...]):
        self._property_changed('classification_constraints')
        self.__classification_constraints = value        

    @property
    def benchmarks(self) -> Tuple[str, ...]:
        """Marquee unique identifiers of assets to be used as benchmarks."""
        return self.__benchmarks

    @benchmarks.setter
    def benchmarks(self, value: Tuple[str, ...]):
        self._property_changed('benchmarks')
        self.__benchmarks = value        

    @property
    def use_machine_learning(self) -> bool:
        """Use Machine learning."""
        return self.__use_machine_learning

    @use_machine_learning.setter
    def use_machine_learning(self, value: bool):
        self._property_changed('use_machine_learning')
        self.__use_machine_learning = value        

    @property
    def lasso_weight(self) -> float:
        """Value of the lasso hyperparameter for machine learning hedges."""
        return self.__lasso_weight

    @lasso_weight.setter
    def lasso_weight(self, value: float):
        self._property_changed('lasso_weight')
        self.__lasso_weight = value        

    @property
    def ridge_weight(self) -> float:
        """Value of the ridge hyperparameter for machine learning hedges"""
        return self.__ridge_weight

    @ridge_weight.setter
    def ridge_weight(self, value: float):
        self._property_changed('ridge_weight')
        self.__ridge_weight = value        


class Hedge(Base):
        
    """Object representation of a Hedge"""

    @camel_case_translate
    def __init__(
        self,
        name: str,
        parameters: dict,
        id_: str = None,
        owner_id: str = None,
        created_by_id: str = None,
        created_time: datetime.datetime = None,
        last_updated_by_id: str = None,
        last_updated_time: datetime.datetime = None,
        entitlements: Entitlements = None,
        entitlement_exclusions: EntitlementExclusions = None,
        tags: Tuple[str, ...] = None,
        description: str = None,
        objective: Union[HedgeObjective, str] = None,
        result: dict = None,
        comparison_results: Tuple[HedgerComparison, ...] = None
    ):        
        super().__init__()
        self.__id = id_
        self.owner_id = owner_id
        self.created_by_id = created_by_id
        self.created_time = created_time
        self.last_updated_by_id = last_updated_by_id
        self.last_updated_time = last_updated_time
        self.entitlements = entitlements
        self.entitlement_exclusions = entitlement_exclusions
        self.tags = tags
        self.name = name
        self.description = description
        self.objective = objective
        self.parameters = parameters
        self.result = result
        self.comparison_results = comparison_results

    @property
    def id(self) -> str:
        """Marquee unique identifier"""
        return self.__id

    @id.setter
    def id(self, value: str):
        self._property_changed('id')
        self.__id = value        

    @property
    def owner_id(self) -> str:
        """Marquee unique identifier for user who owns the object."""
        return self.__owner_id

    @owner_id.setter
    def owner_id(self, value: str):
        self._property_changed('owner_id')
        self.__owner_id = value        

    @property
    def created_by_id(self) -> str:
        """Unique identifier of user who created the object."""
        return self.__created_by_id

    @created_by_id.setter
    def created_by_id(self, value: str):
        self._property_changed('created_by_id')
        self.__created_by_id = value        

    @property
    def created_time(self) -> datetime.datetime:
        """Time created. ISO 8601 formatted string."""
        return self.__created_time

    @created_time.setter
    def created_time(self, value: datetime.datetime):
        self._property_changed('created_time')
        self.__created_time = value        

    @property
    def last_updated_by_id(self) -> str:
        """Unique identifier of user who last updated the object."""
        return self.__last_updated_by_id

    @last_updated_by_id.setter
    def last_updated_by_id(self, value: str):
        self._property_changed('last_updated_by_id')
        self.__last_updated_by_id = value        

    @property
    def last_updated_time(self) -> datetime.datetime:
        """Timestamp of when the object was last updated."""
        return self.__last_updated_time

    @last_updated_time.setter
    def last_updated_time(self, value: datetime.datetime):
        self._property_changed('last_updated_time')
        self.__last_updated_time = value        

    @property
    def entitlements(self) -> Entitlements:
        """Defines the entitlements of a given resource."""
        return self.__entitlements

    @entitlements.setter
    def entitlements(self, value: Entitlements):
        self._property_changed('entitlements')
        self.__entitlements = value        

    @property
    def entitlement_exclusions(self) -> EntitlementExclusions:
        """Defines the exclusion entitlements of a given resource."""
        return self.__entitlement_exclusions

    @entitlement_exclusions.setter
    def entitlement_exclusions(self, value: EntitlementExclusions):
        self._property_changed('entitlement_exclusions')
        self.__entitlement_exclusions = value        

    @property
    def tags(self) -> Tuple[str, ...]:
        """Metadata associated with the object. Provide an array of strings which will be
           indexed for search and locating related objects"""
        return self.__tags

    @tags.setter
    def tags(self, value: Tuple[str, ...]):
        self._property_changed('tags')
        self.__tags = value        

    @property
    def name(self) -> str:
        """Display name of the hedge"""
        return self.__name

    @name.setter
    def name(self, value: str):
        self._property_changed('name')
        self.__name = value        

    @property
    def description(self) -> str:
        """Free text description of a Hedge"""
        return self.__description

    @description.setter
    def description(self, value: str):
        self._property_changed('description')
        self.__description = value        

    @property
    def objective(self) -> Union[HedgeObjective, str]:
        """The objective of the hedge."""
        return self.__objective

    @objective.setter
    def objective(self, value: Union[HedgeObjective, str]):
        self._property_changed('objective')
        self.__objective = get_enum_value(HedgeObjective, value)        

    @property
    def parameters(self) -> dict:
        """The parameters used in the hedge calculation."""
        return self.__parameters

    @parameters.setter
    def parameters(self, value: dict):
        self._property_changed('parameters')
        self.__parameters = value        

    @property
    def result(self) -> dict:
        """Result of the hedge."""
        return self.__result

    @result.setter
    def result(self, value: dict):
        self._property_changed('result')
        self.__result = value        

    @property
    def comparison_results(self) -> Tuple[HedgerComparison, ...]:
        """Results of the comparisons."""
        return self.__comparison_results

    @comparison_results.setter
    def comparison_results(self, value: Tuple[HedgerComparison, ...]):
        self._property_changed('comparison_results')
        self.__comparison_results = value        



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import asyncio
from itertools import chain
import logging
import json
import math
from typing import Iterable, Mapping, Optional, Tuple, Union
from websockets import ConnectionClosedError

from gs_quant.api.risk import RiskApi
from gs_quant.risk import RiskRequest
from gs_quant.session import GsSession


_logger = logging.getLogger(__name__)


class GsRiskApi(RiskApi):

    @classmethod
    def calc_multi(cls, requests: Iterable[RiskRequest]) -> Tuple[Union[dict, str], ...]:
        requests = tuple(requests)
        results = cls._exec(requests)

        if len(results) < len(requests):
            raise RuntimeError('Missing results')

        return tuple(cls.__result(request, result) for request, result in zip(requests, results))

    @classmethod
    def calc(cls, request: RiskRequest) -> Union[Iterable, str]:
        return cls.__result(request, cls._exec(request))

    @classmethod
    def _exec(cls, request: Union[RiskRequest, Iterable[RiskRequest]]) -> Union[Iterable, dict]:
        return GsSession.current._post(cls.__url(request), request, timeout=180)

    @classmethod
    def __result(cls, request: RiskRequest, result: Union[Iterable, dict]) -> Union[dict, str]:
        return cls._handle_results(request, result) if request.waitForResults else result['reportId']

    @classmethod
    def __url(cls, request: Union[RiskRequest, Iterable[RiskRequest]]):
        is_bulk = not isinstance(request, RiskRequest)
        is_internal = any(hasattr(p.instrument, '_derived_type') for p in
                          chain.from_iterable(r.positions for r in (request if is_bulk else (request,))))
        return '/risk{}/calculate{}'.format('-internal' if is_internal else '', '/bulk' if is_bulk else '')

    @classmethod
    def get_results(cls, ids_to_requests: Mapping[str, RiskRequest], timeout: Optional[int] = None)\
            -> Mapping[RiskRequest, Union[Exception, dict]]:
        max_attempts = 5
        send_timeout = 10
        results = {}

        async def get_results():
            error = 'Unknown'
            attempts = 0

            async def impl():
                async with GsSession.current._connect_websocket('/risk/calculate/results/subscribe') as ws:
                    request_ids = [k for k in ids_to_requests.keys() if k not in results]
                    await asyncio.wait_for(ws.send(json.dumps(request_ids)), timeout=send_timeout)

                    while len(results) < len(ids_to_requests):
                        response = await asyncio.wait_for(ws.recv(), timeout=timeout)

                        result_parts = response.split(';')
                        result_id = result_parts[0]
                        raw_result = ';'.join(result_parts[1:])
                        result_str = raw_result[1:]
                        is_error = raw_result.startswith('E')

                        try:
                            result = RuntimeError(result_str) if is_error else\
                                cls._handle_results(ids_to_requests[result_id], json.loads(result_str))
                        except json.decoder.JSONDecodeError:
                            result = RuntimeError(result_str)

                        results[ids_to_requests[result_id]] = result

            while len(results) < len(ids_to_requests) and attempts < max_attempts:
                if attempts > 0:
                    await asyncio.sleep(math.pow(2, attempts))
                    _logger.error('{} error, retrying (attempt {} of {})'.format(error, attempts + 1, max_attempts))

                try:
                    await impl()
                except ConnectionClosedError as cce:
                    error = 'Connection failed: ' + str(cce)
                    attempts += 1
                except asyncio.TimeoutError:
                    error = 'Timed out'
                    attempts = max_attempts
                except Exception as e:
                    error = str(e)
                    attempts = max_attempts

            if len(results) < len(ids_to_requests):
                for request in ids_to_requests.values():
                    if request not in results:
                        results[request] = RuntimeError(error)

        asyncio.run(get_results())

        return results



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import pytest
from pandas.util.testing import assert_series_equal

from gs_quant.timeseries.datetime import *


def test_align():

    dates1 = [
        date(2019, 1, 1),
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 4),
        date(2019, 1, 5),
    ]

    dates2 = [
        date(2019, 1, 2),
        date(2019, 1, 4),
        date(2019, 1, 6),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=dates1)
    y = pd.Series([20.0, 40.0, 60.0], index=dates2)

    expectedl = pd.Series([2.0, 4.0], index=[date(2019, 1, 2), date(2019, 1, 4)])
    expectedr = pd.Series([20.0, 40.0], index=[date(2019, 1, 2), date(2019, 1, 4)])

    result = align(x, y, Interpolate.INTERSECT)
    assert_series_equal(result[0], expectedl, obj="Align intersect left")
    assert_series_equal(result[1], expectedr, obj="Align intersect left")

    result = align(y, x, Interpolate.INTERSECT)
    assert_series_equal(result[0], expectedr, obj="Align intersect right")
    assert_series_equal(result[1], expectedl, obj="Align intersect right")

    union_dates = [
        date(2019, 1, 1),
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 4),
        date(2019, 1, 5),
        date(2019, 1, 6),
    ]

    expected1 = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, np.nan], index=union_dates)
    expected2 = pd.Series([np.nan, 20.0, np.nan, 40.0, np.nan, 60.0], index=union_dates)

    result = align(x, y, Interpolate.NAN)
    assert_series_equal(result[0], expected1, obj="Align NaN left")
    assert_series_equal(result[1], expected2, obj="Align NaN left")

    result = align(y, x, Interpolate.NAN)
    assert_series_equal(result[0], expected2, obj="Align NaN right")
    assert_series_equal(result[1], expected1, obj="Align NaN right")

    expected1 = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, 0.0], index=union_dates)
    expected2 = pd.Series([0.0, 20.0, 0.0, 40.0, 0.0, 60.0], index=union_dates)

    result = align(x, y, Interpolate.ZERO)
    assert_series_equal(result[0], expected1, obj="Align zero left")
    assert_series_equal(result[1], expected2, obj="Align zero left")

    result = align(y, x, Interpolate.ZERO)
    assert_series_equal(result[0], expected2, obj="Align zero right")
    assert_series_equal(result[1], expected1, obj="Align zero right")

    expected1 = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, 5.0], index=union_dates)
    expected2 = pd.Series([20.0, 20.0, 20.0, 40.0, 40.0, 60.0], index=union_dates)

    result = align(x, y, Interpolate.STEP)
    assert_series_equal(result[0], expected1, obj="Align step left")
    assert_series_equal(result[1], expected2, obj="Align step left")

    result = align(y, x, Interpolate.STEP)
    assert_series_equal(result[0], expected2, obj="Align step left")
    assert_series_equal(result[1], expected1, obj="Align step left")

    xp = x.copy()
    yp = y.copy()
    xp.index = pd.to_datetime(xp.index)
    yp.index = pd.to_datetime(yp.index)
    up = pd.to_datetime(union_dates)

    expected1 = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, np.nan], index=up)
    expected2 = pd.Series([np.nan, 20.0, 30.0, 40.0, 50.0, 60.0], index=up)

    result = align(xp, yp, Interpolate.TIME)
    assert_series_equal(result[0], expected1, obj="Align time left")
    assert_series_equal(result[1], expected2, obj="Align time left")

    result = align(yp, xp, Interpolate.TIME)
    assert_series_equal(result[0], expected2, obj="Align time right")
    assert_series_equal(result[1], expected1, obj="Align time right")

    a = pd.Series([0, 100, 110], index=pd.DatetimeIndex(['2019-07-01', '2019-07-08', '2019-07-10']))
    b = pd.Series([20, 60, 70], index=pd.DatetimeIndex(['2019-07-02', '2019-07-10', '2019-07-11']))
    result = align(a, b, Interpolate.TIME)

    u_index = a.index.union(b.index)
    assert_series_equal(result[0], pd.Series([0, 100 / 7, 100, 110, np.nan], index=u_index))
    assert_series_equal(result[1], pd.Series([np.nan, 20, 50, 60, 70], index=u_index))

    result = align(x, 3)
    assert_series_equal(result[0], x, obj="Align scalar left")
    assert_series_equal(result[1], pd.Series(3, index=dates1), obj="Align scalar left")

    result = align(3, x)
    assert_series_equal(result[0], pd.Series(3, index=dates1), obj="Align scalar left")
    assert_series_equal(result[1], x, obj="Align scalar right")

    result = align(1, 2)
    assert result[0] == 1
    assert result[1] == 2

    with pytest.raises(MqValueError):
        align(x, x, "None")


def test_interpolate():

    dates = [
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 5),
        date(2019, 1, 7),
    ]

    x = pd.Series([2.0, 3.0, 5.0, 7.0], index=dates)

    result = interpolate(x, dates)
    assert_series_equal(result, x, obj="Interpolate series by dates")

    result = interpolate(x, x)
    assert_series_equal(result, x, obj="Interpolate series by series dates")

    result = interpolate(x)
    assert_series_equal(result, x, obj="Interpolate series default")

    select_dates = [
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 7),
    ]

    result = interpolate(x, select_dates)
    expected = pd.Series([2.0, 3.0, 7.0], index=select_dates)
    assert_series_equal(result, expected, obj="Interpolate subset of dates")

    select_dates = [
        date(2019, 1, 1),
        date(2019, 1, 2),
        date(2019, 1, 4),
        date(2019, 1, 5),
        date(2019, 1, 6),
        date(2019, 1, 7),
        date(2019, 1, 8),
    ]

    intersect_dates = [
        date(2019, 1, 2),
        date(2019, 1, 5),
        date(2019, 1, 7),
    ]

    result = interpolate(x, select_dates, Interpolate.INTERSECT)
    expected = pd.Series([2.0, 5.0, 7.0], index=intersect_dates)
    assert_series_equal(result, expected, obj="Interpolate intersect")

    result = interpolate(x, select_dates, Interpolate.NAN)
    expected = pd.Series([np.nan, 2.0, np.nan, 5.0, np.nan, 7.0, np.nan], index=select_dates)
    assert_series_equal(result, expected, obj="Interpolate nan")

    result = interpolate(x, select_dates, Interpolate.ZERO)
    expected = pd.Series([0.0, 2.0, 0.0, 5.0, 0.0, 7.0, 0.0], index=select_dates)
    assert_series_equal(result, expected, obj="Interpolate zero")

    result = interpolate(x, select_dates, Interpolate.STEP)
    expected = pd.Series([2.0, 2.0, 2.0, 5.0, 5.0, 7.0, 7.0], index=select_dates)
    assert_series_equal(result, expected, obj="Interpolate step dates")

    result = interpolate(x, pd.Series(np.nan, select_dates), Interpolate.STEP)
    expected = pd.Series([2.0, 2.0, 2.0, 5.0, 5.0, 7.0, 7.0], index=select_dates)
    assert_series_equal(result, expected, obj="Interpolate step series")

    xnan = pd.Series([np.nan, 3.0, 5.0, 7.0], index=dates)

    result = interpolate(xnan, select_dates, Interpolate.STEP)
    expected = pd.Series([np.nan, np.nan, np.nan, 5.0, 5.0, 7.0, 7.0], index=select_dates)
    assert_series_equal(result, expected, obj="Interpolate flat nan start")

    x = pd.Series([2.0, 3.0, 5.0, 7.0], index=pd.DatetimeIndex(dates))
    result = interpolate(x, select_dates, Interpolate.STEP)
    expected = pd.Series([2.0, 2.0, 2.0, 5.0, 5.0, 7.0, 7.0], index=pd.DatetimeIndex(select_dates))
    assert_series_equal(result, expected, obj="Interpolate step dates to series with timestamps")

    with pytest.raises(MqValueError, match="Unknown intersection type: None"):
        interpolate(x, x, "None")

    with pytest.raises(MqValueError, match="Cannot perform step interpolation on an empty series"):
        interpolate(pd.Series(), select_dates, Interpolate.STEP)


def test_value():

    dates = [
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 5),
        date(2019, 1, 7),
    ]

    x = pd.Series([2.0, 3.0, 5.0, 7.0], index=dates)

    result = value(x, date(2019, 1, 3))
    assert result == 3.0

    result = value(x, date(2019, 1, 5))
    assert result == 5.0

    result = value(x, date(2019, 1, 4))
    assert result == 3.0

    result = value(x, date(2019, 1, 4), Interpolate.INTERSECT)
    assert result is None

    result = value(x, date(2019, 1, 4), Interpolate.STEP)
    assert result == 3.0

    result = value(x, date(2019, 1, 4), Interpolate.ZERO)
    assert result == 0.0

    result = value(x, date(2019, 1, 4), Interpolate.NAN)
    assert np.isnan(result)


def test_day():

    dates = [
        date(2019, 1, 1),
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 4),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0], index=dates)

    result = day(x)
    expected = pd.Series([1, 2, 3, 4], index=dates)
    assert_series_equal(result, expected, obj="Day")


def test_weekday():

    dates = [
        date(2019, 1, 7),
        date(2019, 1, 8),
        date(2019, 1, 9),
        date(2019, 1, 10),
        date(2019, 1, 11),
        date(2019, 1, 12),
        date(2019, 1, 13),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0], index=dates)

    result = weekday(x)
    expected = pd.Series([0, 1, 2, 3, 4, 5, 6], index=dates)
    assert_series_equal(result, expected, obj="Weekday")


def test_month():

    dates = [
        date(2019, 1, 1),
        date(2019, 2, 1),
        date(2019, 3, 1),
        date(2019, 4, 1),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0], index=dates)

    result = month(x)
    expected = pd.Series([1, 2, 3, 4], index=dates)
    assert_series_equal(result, expected, obj="Month")


def test_year():

    dates = [
        date(2019, 1, 1),
        date(2020, 1, 2),
        date(2021, 1, 3),
        date(2022, 1, 4),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0], index=dates)

    result = year(x)
    expected = pd.Series([2019, 2020, 2021, 2022], index=dates)
    assert_series_equal(result, expected, obj="Year")


def test_quarter():

    dates = [
        date(2019, 1, 1),
        date(2019, 4, 1),
        date(2019, 7, 1),
        date(2019, 10, 1),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0], index=dates)

    result = quarter(x)
    expected = pd.Series([1, 2, 3, 4], index=dates)
    assert_series_equal(result, expected, obj="Quarter")


def test_day_count_fractions():

    dates = [
        date(2019, 1, 1),
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 4),
        date(2019, 1, 5),
        date(2019, 1, 6),
    ]

    x = pd.Series([])
    assert_series_equal(x, day_count_fractions(x))

    x = pd.Series([100.0, 101, 103.02, 100.9596, 100.9596, 102.978792], index=dates)

    result = day_count_fractions(x, DayCountConvention.ACTUAL_360)
    result2 = day_count_fractions(x.index, DayCountConvention.ACTUAL_360)
    dcf = 1 / 360
    expected = pd.Series([np.NaN, dcf, dcf, dcf, dcf, dcf], index=dates)
    assert_series_equal(result, expected, obj="ACT/360")
    assert_series_equal(result2, expected, obj="ACT/360")

    result = day_count_fractions(x, DayCountConvention.ACTUAL_365F)
    dcf = 1 / 365
    expected = pd.Series([np.NaN, dcf, dcf, dcf, dcf, dcf], index=dates)
    assert_series_equal(result, expected, obj="ACT/365")


def test_date_range():

    dates = [
        date(2019, 1, 1),
        date(2019, 1, 2),
        date(2019, 1, 3),
        date(2019, 1, 4),
        date(2019, 1, 5),
        date(2019, 1, 6),
    ]

    x = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0, 7.0], index=dates)

    assert (date_range(x, 0, 0) == x).all()
    assert (date_range(x, 0, 0, True) == x.iloc[:-2]).all()

    assert date_range(x, 0, date(2019, 1, 3)).index[-1] == date(2019, 1, 3)
    assert (date_range(x, 0, date(2019, 1, 3)) == x.iloc[:3]).all()

    assert date_range(x, date(2019, 1, 3), date(2019, 1, 6)).index[0] == date(2019, 1, 3)
    assert date_range(x, date(2019, 1, 3), date(2019, 1, 6)).index[-1] == date(2019, 1, 6)
    assert (date_range(x, date(2019, 1, 3), date(2019, 1, 6)) == x.iloc[2:6]).all()
